# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Data handler
============

Set of classes that helps developping plugins for ssh and swh. This includes a default
netcdf reader for loading a dataset, and default interpolators for both regular and
irregular grids.
"""
import abc
import datetime
import logging
import os
import re
import time

import dask.array as da
import numba as nb
import numpy as np
import pyinterp
import pyinterp.backends.xarray
import xarray as xr

from . import Interface

LOGGER = logging.getLogger(__name__)


class DatasetLoader:
    """
    Interface that specializes in loading data for the plugin. This is helpful to
    separate the data loading and interpolator definition. The dataset loader has only
    one job : given a time range for interpolation, loads the model dataset that is
    needed to do the said interpolation and transform it to have canonical variable
    names.
    """
    @abc.abstractmethod
    def load_dataset(self, first_date: np.datetime64,
                     last_date: np.datetime64) -> xr.Dataset:
        """
        Loads the data under the form of a xr.Dataset. The loaded dataset should
        contain values that allow interpolating first_date and last_date. This means
        its time interval is a little large than [first_date, last_date].

        Moreover, the dataset should refer to the longitude, latitude, time and sea
        surface height using canonical names: lon, lat, time, ssh

        Args:
            first_date (np.datetime64): The first date that needs to be interpolated
            last_date (np.datetime64): The last date that needs to be interpolated

        Returns:
            xr.Dataset: dataset containing lon, lat, time and ssh variables, with
            canonical names.

        See also:
            _shift_date
        """
        ...

    @staticmethod
    def _shift_date(date: np.datetime64, shift: int,
                    time_delta: np.timedelta64) -> np.datetime64:
        """
        Shift the input date using the time_delta of original data. This is
        useful to generate a time interval for which we need an original value.

        Args:
            date (np.datetime64): interpolation date
            shift (int): 1 for a later date, -1 for an earlier one
            time_delta (np.timedelta64): delta specifying the time resolution of the
            model data.

        Returns:
            The input date if it is the input date is a multiple of time_delta (meaning
            it is on the model time axis). Else, the output is shifted.

        Example:
            If we have data on [t0, t1, dt], and we want an interpolation over [T0, T1],
            then we must make sure that t0 <= T0 - dt and t1 >= T1 + dt. If this
            condition is satisfied, interpolation at T0 and T1 will be possible. If this
            condition is not satisfied, interpolation becomes extrapolation.
        """
        # Before comparing the date and timedelta, ensure they have the same unit
        date_same_unit = date + time_delta - time_delta
        time_delta_same_unit = date + time_delta - date

        if date_same_unit.astype("int64") % time_delta_same_unit.astype(
                "int64") != 0:  # type: ignore
            return date + time_delta * shift
        return date

    @staticmethod
    def _calculate_time_delta(dates: xr.DataArray) -> np.timedelta64:
        """Calculation of the delta T between two consecutive grids."""
        frequency = np.diff(dates)
        try:
            if not np.all(frequency == frequency[0]):
                raise RuntimeError(
                    "Time series does not have a constant step between two "
                    f"grids: {set(frequency)} seconds")
            return np.timedelta64(frequency[0], "ns")
        except IndexError as exc:
            raise RuntimeError(
                "Check that your list of data is not empty") from exc


@nb.njit("(float32[::1])(int64[::1], float32[:, ::1], int64[::1])",
         cache=True,
         nogil=True)  # type:ignore
def _time_interp(xp: np.ndarray, yp: np.ndarray, xi: np.ndarray) -> np.ndarray:
    """Time interpolation for the different spatial grids interpolated on the
    SWOT data"""
    xp_diff = np.diff(xp)

    assert xp.shape[0] == yp.shape[0] and yp.shape[1] == xi.shape[0]
    assert np.all(xp_diff == xp_diff[0])

    result = np.empty(yp.shape[1:], dtype=yp.dtype)

    step = 1.0 / xp_diff[0]
    size = xp.size

    for ix in range(yp.shape[0]):
        index = int(np.around((xi[ix] - xp[0]) * step))
        assert index >= 0 or index <= size
        if index == size - 1:
            i0 = index - 1
            i1 = index
        else:
            i0 = index
            i1 = index + 1
        t0 = xp[i0]
        t1 = xp[i1]
        dt = t1 - t0
        w0 = (t1 - xi[ix]) / dt
        w1 = (xi[ix] - t0) / dt

        for jx in range(yp.shape[1]):
            result[jx] = (w0 * yp[i0, jx] + w1 * yp[i1, jx]) / (w0 + w1)

    return result


class NetcdfLoader(DatasetLoader):
    """
    Plugin that implements a netcdf reader. The netcdf reader works on files whose
    names have the date in it. A pattern (ex. P(?<date>.*).nc), associated with a
    date formatter (ex. %Y%m%d) is used to get build the time series.

    Netcdf files can be expensive to concatenate if there are a lot of files. This
    loader avoid loading too much files by building a dictionary matching file paths
    to their time. During the interpolation, where only a given time period is
    needed, only the files that cover the time period are loaded in the dataset.
    """
    def __init__(self,
                 path: str,
                 date_fmt: str,
                 lon_name: str = "lon",
                 lat_name: str = "lat",
                 ssh_name: str = "ssh",
                 time_name: str = "time",
                 pattern: str = ".nc"):
        """
        Initialization of the netcdf loader.

        Args:
            path (str): Folder containing the netcdf files
            date_fmt (str): date formatter
            lon_name (str): longitude name in the netcdf files. Defaults to
                'lon'
            lat_name (str): latitude name in the netcdf files. Defaults to 'lat'
            ssh_name (str): sea surface height name in the netcdf files.
                Defaults to 'ssh'
            time_name (str): time name in the netcdf files. Defaults to 'time'
            pattern (str): Pattern for the NetCDF file names. It should contain
                the P(?<date>) group to retrieve the time

        Example:
            If we have netcdf files whose names are model_20120305_12h.nc, we must
            define the following to retrieve the time::

                loader = NetcdfLoader(
                    '.',
                    pattern='model_P(?<date>\\w+).nc',
                    date_fmt='%Y%m%d_%Hh'
                )
        """
        if not os.path.exists(path):
            raise FileNotFoundError(f"{path!r}")

        self.lon_name = lon_name
        self.lat_name = lat_name
        self.ssh_name = ssh_name
        self.time_name = time_name

        self.regex = re.compile(pattern).search
        self.date_fmt = date_fmt
        self.time_series = self._walk_netcdf(path)
        self.time_delta = self._calculate_time_delta(self.time_series["date"])

    def _walk_netcdf(self, path: str) -> np.ndarray:
        # Walks a netcdf folder and finds data files in it
        items = []
        length = -1

        for dir_path, _, filenames in os.walk(path):
            for filename in filenames:
                match = self.regex(filename)
                if match:
                    time_counter = np.datetime64(
                        datetime.datetime.strptime(match.group("date"),
                                                   self.date_fmt))
                    filepath = os.path.join(dir_path, filename)
                    items.append((time_counter, filepath))
                    length = max(length, len(filepath))

        # The time series is encoded in a structured Numpy array containing
        # the date and path to the file.
        time_series = np.array(
            items,
            dtype={
                "names": ("date", "path"),
                "formats": ("datetime64[s]", f"U{length}"),
            },
        )
        time_series = time_series[np.argsort(time_series["date"])]
        return time_series

    def select_netcdf_files(self, first_date: np.datetime64,
                            last_date: np.datetime64) -> np.ndarray:
        first_date = self._shift_date(first_date.astype("datetime64[ns]"), -1,
                                      self.time_delta)
        last_date = self._shift_date(last_date.astype("datetime64[ns]"), 1,
                                     self.time_delta)
        if first_date < self.time_series["date"][
                0] or last_date > self.time_series["date"][-1]:
            raise IndexError(
                f"period [{first_date}, {last_date}] is out of range: "
                f"[{self.time_series['date'][0]}, {self.time_series['date'][-1]}]"
            )

        selected = np.logical_and(self.time_series["date"] >= first_date,
                                  self.time_series["date"] < last_date)
        return selected

    def load_dataset(self, first_date: np.datetime64,
                     last_date: np.datetime64):
        LOGGER.debug("fetch data for %s, %s", first_date, last_date)
        selected = self.select_netcdf_files(first_date, last_date)
        dataset = xr.open_mfdataset(self.time_series["path"][selected],
                                    concat_dim=self.time_name,
                                    combine="nested")

        if self.time_name not in dataset.coords:
            LOGGER.debug(
                "Time coordinate %s was not found, assigning "
                "axis with time from file names", self.time_name)
            dataset = dataset.assign_coords(
                {self.time_name: self.time_series["dates"][selected]})

        LOGGER.debug(
            "Renaming dataset with canonical names lon, lat, time, ssh")
        return dataset.rename({
            self.lon_name: "lon",
            self.lat_name: "lat",
            self.ssh_name: "ssh",
            self.time_name: "time"
        })


class IrregularGridHandler(Interface):
    """
    Default interpolator for an irregular grid. First, uses an RTree to do the spatial
    interpolation of all model grid, then do the time interpolation with a simple
    weighting of two grid
    """
    def __init__(self, dataset_loader: DatasetLoader):
        self.dataset_loader = dataset_loader

    def interpolate(self, lon: np.ndarray, lat: np.ndarray, dates: np.ndarray):
        """Interpolate the SSH for the given coordinates"""
        # Spatial interpolation of the SSH on the different selected grids.
        dataset = self.dataset_loader.load_dataset(
            dates.min(),  # type: ignore
            dates.max())  # type: ignore
        dates_p = dataset.time.load().data
        lon_p = dataset.lon.data
        lat_p = dataset.lat.data
        ssh_p = dataset.ssh.data

        start_time = time.time()
        layers = []
        for index in range(len(ssh_p)):
            layers.append(
                self._spatial_interp(ssh_p[index, :], lon_p, lat_p, lon, lat))
        layers = np.stack(layers)
        LOGGER.debug("interpolation completed in %.2fs for period %s, %s",
                     time.time() - start_time, dates.min(), dates.max())

        # Time interpolation of the SSH.
        return _time_interp(
            dates_p.astype("int64"),
            layers,
            dates.astype("datetime64[us]").astype("int64"),
        )

    @staticmethod
    def _spatial_interp(z_model: da.Array, x_model: da.Array,
                        y_model: da.Array, x_sat: np.ndarray,
                        y_sat: np.ndarray) -> np.ndarray:
        mesh = pyinterp.RTree()
        mesh.packing(
            np.vstack((x_model.compute(), y_model.compute())).T,
            z_model.compute())

        z, _ = mesh.radial_basis_function(
            np.vstack((x_sat, y_sat)).T.astype("float32"),
            within=True,
            k=11,
            rbf="thin_plate",
            num_threads=1,
        )
        return z.astype("float32")


class CartesianGridHandler(Interface):
    """
    Default interpolator for regular grid.
    Uses pyinterp.backends.xarray.Grid3D.trivariate interpolator
    """
    def __init__(self, dataset_loader: DatasetLoader):
        self.dataset_loader = dataset_loader

    def interpolate(self, lon: np.ndarray, lat: np.ndarray,
                    dates: np.ndarray) -> np.ndarray:
        """Interpolate the SSH to the required coordinates"""
        dataset = self.dataset_loader.load_dataset(
            dates.min(),  # type: ignore
            dates.max())  # type: ignore

        interpolator = pyinterp.backends.xarray.Grid3D(dataset.ssh)
        ssh = interpolator.trivariate(dict(longitude=lon,
                                           latitude=lat,
                                           time=dates),
                                      bounds_error=True,
                                      interpolator="bilinear")
        return ssh



import hashlib
import urllib.request
import os
import tempfile
import zipfile


class DownloadData:
    """Utility class for downloading test data to the website"""
    FILES = {
        "errors.bin":
        "47a1ef9cb699113ff34f74ac761561ba19f2cb97bfab1896fd28a4c0adce3716",
        "gen_signal_1d.bin":
        "7d18e6dca7e11c7c6be79c5b4c5d6f6cb7e6d6177c43f2e70d4efb143682c037",
        "gen_signal_2d_rectangle.bin":
        "9f8c8578cc5ff40991d4ea0ec2e92858b84c7d52e921345d1428e515a7e5b03c",
    }
    URL = ("https://github.com/CNES/swot_simulator/releases/download"
           "/0.0.0/TestDatasets.zip")

    def __init__(self):
        self.prefix = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   "data")
        os.makedirs(self.prefix, exist_ok=True)
        while not self.check():
            archive = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                                   "TestDatasets.zip")
            try:
                with open(archive, mode="wb") as stream:
                    self.download(stream)
                self.extract(archive)
            finally:
                os.unlink(archive)

    def extract(self, name):
        """Extraction of the downloaded archive."""
        zip_file = zipfile.ZipFile(name)
        for name in zip_file.namelist():
            info = zip_file.getinfo(name)
            deflated = os.path.join(self.prefix, name)
            with open(deflated, "wb") as stream:
                stream.write(zip_file.read(info))

    def check(self):
        """Verify data integrity"""
        for item in self.FILES:
            path = os.path.join(self.prefix, item)
            if not os.path.exists(path) or \
                    self.sha256sum(path) != self.FILES[item]:
                return False
        return True

    @classmethod
    def download(cls, stream):
        """Download data from bitbucket"""
        response = urllib.request.urlopen(cls.URL)

        while True:
            data = response.read(65536)
            if not data:
                break
            stream.write(data)

    @staticmethod
    def sha256sum(path):
        """
        Computes the SHA256 hash for a file
        """
        sha256 = hashlib.sha256()
        with open(path, 'rb') as stream:
            for block in iter(lambda: stream.read(65536), b''):
                sha256.update(block)
        return sha256.hexdigest()


DownloadData()


import os
import pickle
import numpy as np
import xarray as xr

import swot_simulator
import swot_simulator.random_signal as random_signal

ROOT = os.path.dirname(os.path.abspath(__file__))


def test_gen_signal_1d():
    with open(os.path.join(ROOT, "data", "gen_signal_1d.bin"), "rb") as stream:
        (fi, psi, x, nseed, fmin, fmax, alpha, lf_extpl, hf_extpl,
         _expected) = pickle.load(stream)

    rng = np.random.default_rng(seed=nseed)
    result = random_signal.gen_signal_1d(fi, psi, x, rng, fmin, fmax, alpha,
                                         lf_extpl, hf_extpl)
    assert result.mean() < 1


def test_gen_signal_2d_rectangle():
    with open(os.path.join(ROOT, "data", "gen_signal_2d_rectangle.bin"),
              "rb") as stream:
        (fi, psi, x, y, fminx, fminy, fmax, alpha, nseed, lf_extpl, hf_extpl,
         _expected) = pickle.load(stream)

    ps2d, f = random_signal.gen_ps2d(fi, psi, fminx, fminy, fmax, alpha,
                                     lf_extpl, hf_extpl)
    rng = np.random.default_rng(seed=nseed)
    result = random_signal.gen_signal_2d_rectangle(ps2d, f, x, y, rng, fminx,
                                                   fminy, fmax, alpha)
    assert result.mean() < 1


def test_read_file_karin():
    height_sdt, cross_track, swh = random_signal.read_file_karin(
        str(swot_simulator.DATA.joinpath("karin_noise_v2.nc")))


def test_read_file_instr():
    dataset = random_signal.read_file_instr(
        str(swot_simulator.DATA.joinpath("error_spectrum.nc")), 2.0)
    assert isinstance(dataset, xr.Dataset)



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Estimated roll errors
---------------------
"""
from typing import Dict, Tuple
import numpy as np
import xarray as xr

from .. import settings


class CorrectedRollPhase:
    """
    Corrected roll errors

    Args:
        parameters (settings.Parameters): Simulation settings
        first_date (numpy.datetime64): Date of the first simulated
            measurement.
    """
    def __init__(self, parameters: settings.Parameters,
                 first_date: np.datetime64) -> None:
        with xr.open_dataset(parameters.corrected_roll_phase_dataset) as ds:
            time_date = first_date + (ds.time[:].astype(np.float32) * 86400 *
                                      1000000).astype("timedelta64[us]")
            self.time_date = time_date.astype('datetime64[us]').astype(
                'float64') * 0.001
            self.proll_err = ds.proll_err
            self.p1phase_err = ds.p1phase_err
            self.p2phase_err = ds.p2phase_err
            self.slope1_est = ds.slope1_est
            self.slope2_est = ds.slope2_est
            self.slope1_err = ds.slope1_err
            self.slope2_err = ds.slope2_err

    def _generate_1d(
            self,
            time: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        time = time.astype('datetime64[us]').astype('float64')
        phase_l = np.interp(time, self.time_date, self.p1phase_err)
        phase_r = np.interp(time, self.time_date, self.p2phase_err)
        phase1d = np.concatenate(([phase_l.T], [phase_r.T]), axis=0)
        roll = np.interp(time, self.time_date, self.proll_err)

        est_l = np.interp(time, self.time_date, self.slope1_est)
        err_l = np.interp(time, self.time_date, self.slope1_err)
        rem_l = est_l - err_l
        est_r = np.interp(time, self.time_date, self.slope2_est)
        err_r = np.interp(time, self.time_date, self.slope2_err)
        rem_r = est_r - err_r
        theta2 = np.concatenate(([rem_l.T], [rem_r.T]), axis=0)

        return roll * 1e-3, phase1d.T * 1e-3, theta2.T * 1e-3

    def generate(
        self,
        time: np.ndarray,
        x_ac: np.ndarray,
    ) -> Dict[str, np.ndarray]:
        """Interpolate roll and phase and errors

        Args:
            time (numpy.ndarray): Date of measurements.
            x_ac (numpy.ndarray): Across track distance.

        Returns:
            dict: variable name and errors simulated.
        """
        roll_1d, phase_1d, rollphase_est_1d = self._generate_1d(time)
        num_pixels = x_ac.shape[0]
        swath_center = num_pixels // 2
        ac_l = x_ac[:swath_center]
        ac_r = x_ac[swath_center:]

        phase = np.full((phase_1d.shape[0], num_pixels), np.nan)
        phase[:, :swath_center] = ac_l * phase_1d[:, 0, np.newaxis]
        phase[:, swath_center:] = ac_r * phase_1d[:, 1, np.newaxis]

        rollphase_est = np.full((phase_1d.shape[0], num_pixels), np.nan)
        rollphase_est[:, :swath_center] = np.mat(rollphase_est_1d[:,
                                                                  0]).T * ac_l
        rollphase_est[:,
                      swath_center:] = np.mat(rollphase_est_1d[:, 1]).T * ac_r
        return {
            "simulated_error_roll": x_ac * roll_1d[:, np.newaxis],
            "simulated_error_phase": phase,
            "simulated_roll_phase_estimate": rollphase_est
        }



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Interpolation of the SWH from WW3
=================================
"""
import numpy as np
import pyinterp.backends.xarray

from .. import data_handler


class WW3(data_handler.CartesianGridHandler):
    """
    Interpolation of the SWH from WW3.
    """
    def __init__(self, path: str):
        loader = data_handler.NetcdfLoader(path,
                                           date_fmt="%Y%m%d",
                                           ssh_name="hs",
                                           pattern=r"ww3.(?P<date>\w+)_hs\.nc")
        super().__init__(loader)

    def interpolate(self, lon: np.ndarray, lat: np.ndarray,
                    dates: np.ndarray) -> np.ndarray:
        """Interpolate the SSH to the required coordinates"""
        dataset = self.dataset_loader.load_dataset(
            dates.min(),  # type: ignore
            dates.max())  # type: ignore

        interpolator = pyinterp.backends.xarray.Grid3D(dataset.ssh)
        swh = interpolator.trivariate(
            dict(longitude=lon, latitude=lat, time=dates),
            interpolator="bilinear",
        )
        return swh



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Interpolation of the SSH HYCOM
==============================
"""
import numpy as np

from .. import data_handler


class HYCOM(data_handler.CartesianGridHandler):
    """
    Interpolation of the SSH HYCOM
    """
    def __init__(self, path: str):
        #: Decode the product date encoded from the file name.
        # hycom_GLBu0.08_191_2012031900_t012.nc
        loader = HYCOM.OverriddenNetcdfLoader(
            path,
            date_fmt="%Y%m%d",
            ssh_name="surf_el",
            pattern=r"hycom_GLBu0.08_191_(?P<date>\d{8})\d{2}_t\d{3}.nc")
        super().__init__(loader)

    class OverriddenNetcdfLoader(data_handler.NetcdfLoader):
        def load_dataset(self, first_date: np.datetime64,
                         last_date: np.datetime64):
            dataset = super().load_dataset(first_date, last_date)
            return dataset.sel(depth=0)



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Write SWOT data to netCDF4 files.
=================================
"""
from typing import Any, Dict, Hashable, List, Optional, Tuple, Union
import collections
import copy
import datetime
import logging
import os
import pathlib

#
import netCDF4
import numpy as np
import xarray as xr

#
from . import __version__
from . import math
from . import orbit_propagator
from . import product_specification

#: Logger for this module
LOGGER = logging.getLogger(__name__)

#: The reference document of this simulator.
REFERENCE = "Gaultier, L., C. Ubelmann, and L.-L. Fu, 2016: The " \
    "Challenge of Using Future SWOT Data for Oceanic Field Reconstruction." \
    " J. Atmos. Oceanic Technol., 33, 119-126, doi:10.1175/jtech-d-15-0160" \
    ".1. http://dx.doi.org/10.1175/JTECH-D-15-0160.1."

#: Product specification file for the nadir product
NADIR_SPECIFICATION = pathlib.Path(__file__).parent / "nadir-gpr.xml"


def global_attributes(attributes: Dict[str, Dict[str, Any]], cycle_number: int,
                      pass_number: int, date: np.ndarray, lng: np.ndarray,
                      lat: np.ndarray, lon_eq: float,
                      time_eq: np.datetime64) -> Dict[str, Any]:
    """Calculates the global attributes of the pass"""
    def _iso_date(date: np.datetime64) -> str:
        """Return the time formatted according to ISO."""
        epoch = date.astype("datetime64[us]").astype(
            "int64") * 1e-6  # type: ignore
        return datetime.datetime.utcfromtimestamp(epoch).isoformat() + "Z"

    now = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%SZ : Creation")

    ellipsoid_semi_major_axis = product_specification.cast_to_dtype(
        6378137, attributes["ellipsoid_semi_major_axis"])
    ellipsoid_flattening = product_specification.cast_to_dtype(
        1 / 298.25722356, attributes["ellipsoid_flattening"])
    result = collections.OrderedDict(
        Conventions="CF-1.7",
        title=attributes["title"]["attrs"]["description"],
        institution="CNES/JPL",
        source="Simulate product",
        history=now,
        platform="SWOT",
        product_version=__version__,
        references=REFERENCE,
        reference_document="D-56407_SWOT_Product_Description_L2_LR_SSH",
        contact="CNES aviso@altimetry.fr, JPL podaac@podaac.jpl.nasa.gov",
        cycle_number=product_specification.cast_to_dtype(
            cycle_number, attributes["cycle_number"]),
        pass_number=product_specification.cast_to_dtype(
            pass_number, attributes["pass_number"]),
        equator_time=_iso_date(time_eq),
        equator_longitude=math.normalize_longitude(lon_eq, 0),  # type: ignore
        time_coverage_start=_iso_date(date[0]),
        time_coverage_end=_iso_date(date[-1]),
        geospatial_lon_min=lng.min(),
        geospatial_lon_max=lng.max(),
        geospatial_lat_min=lat.min(),
        geospatial_lat_max=lat.max())
    if len(lng.shape) == 2:
        result.update([
            ("left_first_longitude", lng[0, 0]),
            ("left_first_latitude", lat[0, 0]),
            ("left_last_longitude", lng[-1, 0]),
            ("left_last_latitude", lat[-1, 0]),
            ("right_first_longitude", lng[0, -1]),
            ("right_first_latitude", lat[0, -1]),
            ("right_last_longitude", lng[-1, -1]),
            ("right_last_latitude", lat[-1, -1]),
        ])
    result.update([
        ("wavelength",
         product_specification.cast_to_dtype(0.008385803020979,
                                             attributes["wavelength"])),
        ("orbit_solution", "POE")
    ])
    for item in attributes:
        if item.startswith("xref_input"):
            result[item] = "N/A"
    result.update([("ellipsoid_semi_major_axis", ellipsoid_semi_major_axis),
                   ("ellipsoid_flattening", ellipsoid_flattening)])
    return result


def _split_group_name(name):
    """Get the name of the group and the variable deduced from the name."""
    if "/" in name:
        return tuple(name.rsplit("/", 1))
    return None, name


def _create_variable_args(
        encoding: Dict[Hashable, Dict], name: Hashable,
        variable: xr.DataArray) -> Tuple[Hashable, Dict[str, Any]]:
    """Initiation of netCDF4.Dataset.createVariable method parameters from
    user-defined encoding information.
    """
    kwargs = dict()
    keywords = encoding[name] if name in encoding else dict()
    if "_FillValue" in keywords:
        keywords["fill_value"] = keywords.pop("_FillValue")
    dtype = keywords.pop("dtype", variable.dtype)
    for key, value in dict(zlib=True,
                           complevel=4,
                           shuffle=True,
                           fletcher32=False,
                           contiguous=False,
                           chunksizes=None,
                           endian='native',
                           least_significant_digit=None,
                           fill_value=None).items():
        kwargs[key] = keywords.pop(key, value)
    return dtype, kwargs


def _group_attributes(side: product_specification.Side) -> Dict[str, str]:
    """Gets the attributes of a group (unsmoothed products)."""
    if side == product_specification.Side.LEFT.value:
        return dict(description="Unsmoothed SSH measurement data and related "
                    "information for the left half swath.")
    return dict(description="Unsmoothed SSH measurement data and related "
                "information for the right half swath.")


def _create_variable(xr_dataset: xr.Dataset, nc_dataset: netCDF4.Dataset,
                     encoding: Dict[str, Dict[str, Dict[str, Any]]],
                     name: Hashable, unlimited_dims: Optional[List[str]],
                     variable: xr.DataArray) -> None:
    """Creation and writing of the NetCDF variable"""
    unlimited_dims = unlimited_dims or list()

    variable.attrs.pop("_FillValue", None)
    # Encode datetime64 to float64
    if np.issubdtype(variable.dtype, np.datetime64):
        # 946684800000000 number of microseconds between 2000-01-01 and
        # 1970-01-01
        values = (variable.values.astype("datetime64[us]").astype("int64") -
                  946684800000000) * 1e-6
        if variable.name in xr_dataset.coords:
            xr_dataset = xr_dataset.assign_coords(
                coords={variable.name: values})
            attrs = variable.attrs
            variable = xr_dataset[variable.name]
            variable.attrs.update(attrs)
        else:
            variable.values = values
        assert (
            variable.attrs["units"] == "seconds since 2000-01-01 00:00:00.0")
    dtype, kwargs = _create_variable_args(encoding, name, variable)

    group, name = _split_group_name(name)

    if group is not None:
        if group not in nc_dataset.groups:
            nc_dataset = nc_dataset.createGroup(group)
            if group in ["left", "right"]:
                nc_dataset.setncatts(
                    _group_attributes(
                        getattr(product_specification.Side,
                                group.upper()).value))
        else:
            nc_dataset = nc_dataset.groups[group]

    # If the dimensions doesn't exist then we have to create them.
    if not nc_dataset.dimensions:
        for dim_name, size in xr_dataset.dims.items():
            dim_group, dim_name = _split_group_name(dim_name)
            if dim_group == group:
                nc_dataset.createDimension(
                    dim_name, None if dim_name in unlimited_dims else size)

    ncvar = nc_dataset.createVariable(
        name, dtype,
        tuple(_split_group_name(item)[-1] for item in variable.dims), **kwargs)
    ncvar.setncatts(variable.attrs)
    values = variable.values
    if kwargs['fill_value'] is not None:
        if values.dtype.kind == "f" and np.any(np.isnan(values)):
            values[np.isnan(values)] = kwargs['fill_value']
        values = np.ma.array(values, mask=values == kwargs['fill_value'])
    nc_dataset[name][:] = values


def to_netcdf(dataset: xr.Dataset,
              path: Union[str, pathlib.Path],
              encoding: Optional[Dict[str, Dict]] = None,
              unlimited_dims: Optional[List[str]] = None,
              **kwargs):
    """Write dataset contents to a netCDF file"""
    encoding = encoding or dict()

    if isinstance(path, str):
        path = pathlib.Path(path)
    path.parent.mkdir(parents=True, exist_ok=True)

    with netCDF4.Dataset(path, **kwargs) as stream:
        stream.setncatts(dataset.attrs)

        for name, variable in dataset.coords.items():
            _create_variable(dataset, stream, encoding, name, unlimited_dims,
                             variable)

        for name, variable in dataset.data_vars.items():
            _create_variable(dataset, stream, encoding, name, unlimited_dims,
                             variable)


def _write_nadir_product(ds: xr.Dataset, path: str,
                         complete_product: bool) -> None:
    """Write the nadir product to a netCDF file."""
    variables, _attributes = product_specification.parse_specification_file(
        str(NADIR_SPECIFICATION))
    variables["data_01/ku/simulated_error_altimeter"] = dict(
        attrs=dict(_FillValue=2147483647,
                   long_name='Altimeter error',
                   standard_name='',
                   units='m',
                   scale_factor=0.0001,
                   coordinates='/data_01/longitude /data_01/latitude'),
        dtype='int32',
        shape=("data_01/time", ))

    ds = ds.rename_dims({"num_lines": "data_01/time"})
    ds = ds.rename_vars(
        dict(time="data_01/time",
             latitude_nadir="data_01/latitude",
             longitude_nadir="data_01/longitude",
             simulated_error_altimeter="data_01/ku/simulated_error_altimeter"))
    # Rename the simulated variables to match the specification pattern (These
    # variables do not exist in the official product. Only the SSHA is present.)
    for name in ["simulated_true_ssh_nadir", "ssh_nadir", "swh_nadir"]:
        if name in ds.variables:
            group_path = f"data_01/ku/{name.replace('_nadir', '')}"
            ds = ds.rename_vars({name: group_path})
            attrs = ds.variables[group_path].attrs.copy()
            attrs["coordinates"] = "/data_01/longitude /data_01/latitude"
            attrs["_FillValue"] = 2147483647
            variables[group_path] = dict(attrs=attrs,
                                         dtype='int32',
                                         shape=("data_01/time", ))
    data_vars = {}
    encoding = {}
    for name in variables:
        if name not in ds.data_vars:
            continue
        encoding_array, array = product_specification.build_array(
            name, variables, ds.variables[name].values)
        data_vars[array.name] = array
        encoding.update(encoding_array)

    shape = ds.variables["data_01/time"].shape

    if complete_product:
        for item, properties in variables.items():
            if item in data_vars:
                continue
            data = np.full(shape,
                           product_specification.encode_fill_value(properties),
                           properties["dtype"])
            encoding_array, array = product_specification.build_array(
                item, variables, data)
            data_vars[array.name] = array
            encoding.update(encoding_array)

    # import pdb; pdb.set_trace()
    ds.attrs["title"] = "GDR - Reduced dataset"
    ds = xr.Dataset(data_vars=data_vars, attrs=ds.attrs)
    to_netcdf(ds, path, encoding=encoding, mode="w")


class Nadir:
    """Handle the nadir measurements dataset.

    Args:
        track (orbit_propagator.Pass): Properties of the half-orbit to write
    """
    def __init__(self,
                 track: orbit_propagator.Pass,
                 product_type: Optional[str] = None):
        self.product_spec = product_specification.ProductSpecification(
            product_type)
        self.num_lines = track.time.size
        self.num_pixels = 1
        self.encoding, self.data_vars = self.product_spec.time(track.time)
        self._data_array("lon_nadir", track.lon_nadir)
        self._data_array("lat_nadir", track.lat_nadir)
        self._time_eq = track.time_eq
        self._lon_eq = track.lon_eq
        self.ascending = np.all(np.diff(track.lat_nadir) > 0)

    def _data_array(self,
                    attr: str,
                    data: np.ndarray,
                    fill_value: Optional[np.ndarray] = None) -> None:
        """Get the properties of a data array to be inserted in the dataset.

        Args:
            attr (dict): Name of the method of the `ProductSpec` class defining
                the variable to be inserted.
            data (np.array): data to be recorded
            fill_value (np.array, optional): Fill value of the reference
                track vector to be inserted in the centre of the swath.
        """
        # Is it necessary to insert a central pixel?
        if len(data.shape) == 2 and self.num_pixels % 2 == 1:
            # If the data is a grid, then the defined fill values are inserted
            # or a Nan vector if the fill values are undefined.
            if fill_value is None:
                fill_value = np.full((data.shape[0], ),
                                     np.nan,
                                     dtype=np.float64)
            middle = data.shape[1] // 2
            data = np.c_[data[:, :middle], fill_value[:, np.newaxis],
                         data[:, middle:]]
        variable = getattr(self.product_spec, attr)(data)
        if variable is not None:
            encoding, array = variable
            self.encoding.update(encoding)
            self.data_vars += array

    def ssh(self, array: np.ndarray) -> None:
        """Sets the variable describing the SSH to nadir.

        Args:
            array (np.ndarray): Data to be recorded
        """
        self._data_array("ssh_nadir", array)

    def swh(self, array: np.ndarray) -> None:
        """Sets the variable describing the SWH to nadir.

        Args:
            array (np.ndarray): Data to be recorded
        """
        self._data_array("swh_nadir", array)

    def simulated_true_ssh(self, array: np.ndarray) -> None:
        """Sets the variable describing the SSH to nadir free of measurement
        errors.

        Args:
            array (np.ndarray): Data to be recorded
        """
        self._data_array("simulated_true_ssh_nadir", array)

    def update_noise_errors(self, noise_errors: Dict[str, np.ndarray]) -> None:
        """Sets the values of the simulated errors.

        Args:
            noise_errors (dict): Simulated errors to be recorded
        """
        for k, v in noise_errors.items():
            if v.ndim == 1:
                self._data_array(k, v)

    def to_xarray(self, cycle_number: int, pass_number: int,
                  complete_product: bool) -> xr.Dataset:
        """Converts this instance into a xarray dataset.

        Args:
            cycle_number (int): Cycle number.
            pass_number (int): Pass number.
            complete_product (bool): True if you want to obtain a complete
                SWOT dataset, i.e. containing all the variables of the
                official dataset, even those not calculated by the simulator.

        Returns:
            xr.Dataset: xarray dataset
        """
        data_vars = dict((item.name, item) for item in self.data_vars)

        if "longitude" in data_vars:
            lng = data_vars["longitude"]
            lat = data_vars["latitude"]
        elif f"{product_specification.Side.LEFT.value}/longitude" in data_vars:
            lng = np.c_[
                data_vars[f"{product_specification.Side.LEFT.value}/longitude"],
                data_vars[
                    f"{product_specification.Side.RIGHT.value}/longitude"]]
            lat = np.c_[
                data_vars[f"{product_specification.Side.LEFT.value}/latitude"],
                data_vars[f"{product_specification.Side.RIGHT.value}/latitude"]]
        else:
            lng = data_vars["longitude_nadir"]
            lat = data_vars["latitude_nadir"]

        # Variables that are not calculated are filled in in order to have a
        # product compatible with the PDD SWOT. Longitude is used as a
        # template.
        pdd_vars = copy.copy(self.data_vars)
        if complete_product and len(lng.shape) == 2:

            # Creation of an associative dictionary between the name of the
            # dimensions of the dataset and their values.
            dimensions = []
            {
                dimensions.extend(list(zip(item.dims, item.shape)))
                for item in data_vars.values()
            }
            shape = dict(dimensions)

            # The simulator does not handle variables using the "num_sides"
            # dimension.
            shape.update(self.product_spec.num_sides())
            for encoding, array in self.product_spec.fill_variables(
                    data_vars.keys(), shape):
                self.encoding.update(encoding)
                pdd_vars += array

        # Variables must be written in the declaration order of the XML file
        unordered_vars = dict((item.name, item) for item in pdd_vars)
        data_vars = collections.OrderedDict()
        for item in self.product_spec.variables:
            if item in unordered_vars:
                data_vars[item] = unordered_vars[item]

        # Add variables that are not defined in the XML file
        for item in set(unordered_vars.keys()) - set(data_vars.keys()):
            data_vars[item] = unordered_vars[item]

        return xr.Dataset(data_vars=data_vars,
                          attrs=global_attributes(self.product_spec.attributes,
                                                  cycle_number, pass_number,
                                                  self.data_vars[0].values,
                                                  np.asarray(lng),
                                                  np.asarray(lat),
                                                  self._lon_eq, self._time_eq))

    def _write_netcdf(self, path: str, cycle_number: int, pass_number: int,
                      complete_product: bool) -> None:
        """Writes the Nadir dataset to a netCDF file.

        This method must be specialized to write Nadir and Karin products.
        Indeed, these two products have different structures and cannot be
        processed in the same way.
        """
        dataset = self.to_xarray(cycle_number, pass_number, False)
        _write_nadir_product(dataset, path, complete_product)

    def to_netcdf(self, cycle_number: int, pass_number: int, path: str,
                  complete_product: bool) -> None:
        """Writes the dataset in a netCDF file.

        Args:
            cycle_number (int): Cycle number.
            pass_number (int): Pass number.
            complete_product (bool): True if you want to obtain a complete
                SWOT dataset, i.e. containing all the variables of the
                official dataset, even those not calculated by the simulator.
        """
        LOGGER.info("write %s", path)
        try:
            self._write_netcdf(path, cycle_number, pass_number,
                               complete_product)
        except:
            if os.path.exists(path):
                os.unlink(path)
            raise


class Swath(Nadir):
    """Handle the KaRIn measurements dataset.

    Args:
        track (orbit_propagator.Pass): Properties of the half-orbit to write
        central_pixel (bool): Inserts or not in the swath, a central pixel
            divided in two by the reference ground track.
    """
    def __init__(self,
                 track: orbit_propagator.Pass,
                 central_pixel: bool = False,
                 product_type: Optional[str] = None) -> None:
        super().__init__(track, product_type)
        self.num_pixels = track.x_ac.size + int(central_pixel)
        _x_ac = np.full((track.time.size, track.x_ac.size),
                        track.x_ac,
                        dtype=track.x_ac.dtype)
        self._data_array("x_ac", _x_ac,
                         np.full((track.time.size, ), 0, dtype=np.float64))
        self._data_array("lon", track.lon, track.lon_nadir)
        self._data_array("lat", track.lat, track.lat_nadir)

    def ssh(self, array: np.ndarray) -> None:
        """Sets the variable describing the KaRIn SSH.

        Args:
            array (np.ndarray): Data to be recorded
        """
        self._data_array("ssh_karin", array)

    def swh(self, array: np.ndarray) -> None:
        """Sets the variable describing the KaRIn SWH.

        Args:
            array (np.ndarray): Data to be recorded
        """
        if not self.product_spec.unsmoothed:
            self._data_array("swh_karin", array)

    def simulated_true_ssh(self, array: np.ndarray) -> None:
        """Sets the variable describing the KaRIn SSH free of measurement
        errors.

        Args:
            array (np.ndarray): Data to be recorded
        """
        self._data_array("simulated_true_ssh_karin", array)

    def update_noise_errors(self, noise_errors: Dict[str, np.ndarray]) -> None:
        """Sets the values of the simulated errors.

        Args:
            noise_errors (dict): Simulated errors to be recorded
        """
        if self.product_spec.ssh_karin_name() in self.product_spec:
            for k, v in noise_errors.items():
                if v.ndim == 2:
                    self._data_array(k, v)

    def _write_netcdf(self, path: str, cycle_number: int, pass_number: int,
                      complete_product: bool) -> None:
        """Writes the swath dataset to a netCDF file."""
        dataset = self.to_xarray(cycle_number, pass_number, complete_product)
        to_netcdf(dataset, path, self.encoding, mode="w")



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Mathematical routines
---------------------
"""
from typing import Optional, Tuple
import collections
import numpy as np
import numba as nb


@nb.njit(cache=True, nogil=True)
def normalize_longitude(lon: np.ndarray,
                        lon_min: Optional[float] = -180.0) -> np.ndarray:
    """Normalize longitudes between [lon_min, lon_min + 360]"""
    return ((lon - lon_min) % 360) + lon_min


@nb.njit('UniTuple(float64[:], 3)(float64[:], float64[:])',
         cache=True,
         nogil=True)
def spher2cart(lon: np.ndarray,
               lat: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Convert spherical coordinates to cartesian coordinates"""
    rlon = np.radians(lon)
    rlat = np.radians(lat)
    x = np.cos(rlon) * np.cos(rlat)
    y = np.sin(rlon) * np.cos(rlat)
    z = np.sin(rlat)
    return x, y, z


@nb.njit('UniTuple(float64[:], 2)(float64[:], float64[:], float64[:])',
         cache=True,
         nogil=True)
def cart2spher(x: np.ndarray, y: np.ndarray,
               z: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Convert cartesian coordinates to spherical coordinates"""
    indexes = np.where((x == 0) & (y == 0))[0]
    if indexes.size:
        x[indexes] = np.nan
        y[indexes] = np.nan
    lat = np.arctan2(z, np.sqrt(x * x + y * y))
    lon = np.arctan2(y, x)
    if indexes.size:
        lon[indexes] = 0
        lat[indexes] = np.pi * 0.5 * np.sign(z[indexes])
    return np.degrees(lon), np.degrees(lat)


@nb.njit('float64[:, ::1](float64, float64[::1])', cache=True, nogil=True)
def rotation_3d_matrix(theta: float, axis: np.ndarray) -> np.ndarray:
    """Creates a rotation matrix: Slow method.

    Inputs are rotation angle theta and rotation axis axis.
    The rotation matrix correspond to a rotation of angle theta
    with respect to axis axis."""
    axis = axis / np.sqrt(np.dot(axis, axis))
    a = np.cos(theta * 0.5)
    b, c, d = -axis * np.sin(theta * 0.5)
    result = np.empty((3, 3), dtype=axis.dtype)
    result[0, 0] = a * a + b * b - c * c - d * d
    result[0, 1] = 2 * (b * c - a * d)
    result[0, 2] = 2 * (b * d + a * c)
    result[1, 0] = 2 * (b * c + a * d)
    result[1, 1] = a * a + c * c - b * b - d * d
    result[1, 2] = 2 * (c * d - a * b)
    result[2, 0] = 2 * (b * d - a * c)
    result[2, 1] = 2 * (c * d + a * b)
    result[2, 2] = a * a + d * d - b * b - c * c
    return result


@nb.njit('float64[::1](float64[::1], float64[::1], float64)',
         cache=True,
         nogil=True)
def curvilinear_distance(lon: np.ndarray, lat: np.ndarray,
                         radius: float) -> np.ndarray:
    """Calculating the curvilinear distance."""
    lat1 = np.radians(lat[0:-1])
    lat2 = np.radians(lat[1:])
    dlon = np.radians(np.diff(lon))

    distance = np.zeros_like(lon)
    distance[1:] = np.arccos(
        np.sin(lat1) * np.sin(lat2) +
        np.cos(lat1) * np.cos(lat2) * np.cos(dlon)) * radius
    return np.cumsum(distance)


@nb.njit('float64[:, :](float64[:, :])', cache=True, nogil=True)
def satellite_direction(location: np.ndarray) -> np.ndarray:
    """Calculate satellite direction"""
    direction = np.empty_like(location)

    denominator = np.sqrt(location[1:-1, 0]**2 + location[1:-1, 1]**2 +
                          location[1:-1, 2]**2)
    direction[1:-1, 0] = (location[2:, 0] - location[:-2, 0]) / denominator
    direction[1:-1, 1] = (location[2:, 1] - location[:-2, 1]) / denominator
    direction[1:-1, 2] = (location[2:, 2] - location[:-2, 2]) / denominator

    direction[0, :] = direction[1, :]
    direction[0, :] = direction[1, :]
    direction[0, :] = direction[1, :]
    direction[-1, :] = direction[-2, :]
    direction[-1, :] = direction[-2, :]
    direction[-1, :] = direction[-2, :]

    return direction


@nb.njit('UniTuple(float64, 2)(float64, float64, float64)',
         cache=True,
         nogil=True)
def __cartesian2spherical(x: float, y: float, z: float) -> Tuple[float, float]:
    """Convert cartesian coordinates to spherical coordinates for scalar
    values. This function is for internal use only"""
    if x == 0 and y == 0:
        return 0, np.degrees(np.pi * 0.5 * np.sign(z))
    lat = np.arctan2(z, np.sqrt(x * x + y * y))
    lon = np.arctan2(y, x)
    return np.degrees(lon), np.degrees(lat)


@nb.njit('UniTuple(float64[:, ::1], 2)(float64, float64, int64, float64, '
         'float64[:, ::1], float64[:, ::1])',
         cache=True,
         nogil=True)
def calculate_swath(delta_ac: float, half_gap: float, half_swath: int,
                    radius: float, location: np.ndarray,
                    direction: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """Calculation of swath geometry"""
    lon = np.empty((location.shape[0], 2 * half_swath), dtype=np.float64)
    lat = np.empty((location.shape[0], 2 * half_swath), dtype=np.float64)

    for ix in range(len(location)):
        for jx in range(0, int(half_swath)):
            rotation = rotation_3d_matrix(-(jx * delta_ac + half_gap) / radius,
                                          direction[ix, :])

            loc = np.dot(rotation, location[ix])
            kx = half_swath + jx
            lon[ix,
                kx], lat[ix,
                         kx] = __cartesian2spherical(loc[0], loc[1], loc[2])

            loc = np.dot(rotation.T, location[ix])
            kx = half_swath - jx - 1
            lon[ix,
                kx], lat[ix,
                         kx] = __cartesian2spherical(loc[0], loc[1], loc[2])
    return lon, lat


Point = collections.namedtuple('Point', ['lon', 'lat'])
Point.__doc__ = """Defines a 2D Point"""


class Box:
    """Defines a box made of two describing points.
    """
    def __init__(self,
                 min_corner: Optional[Point] = None,
                 max_corner: Optional[Point] = None):
        self.min_corner = min_corner or Point(-180, -90)
        self.max_corner = max_corner or Point(180, 90)

    def __repr__(self):
        return "%s(%s, %s)" % (self.__class__.__name__, str(
            self.min_corner), str(self.max_corner))

    def within(self, lon: np.ndarray, lat: np.ndarray) -> np.ndarray:
        """Returns true if the coordinates are located in the box."""
        lon = normalize_longitude(lon, lon_min=self.min_corner.lon)
        lon_is_in_range = (lon >= self.min_corner.lon) | (
            lon <= self.max_corner.lon
        ) if self.max_corner.lon < self.min_corner.lon else (
            lon >= self.min_corner.lon) & (lon <= self.max_corner.lon)

        return (lat >= self.min_corner.lat) & (
            lat <= self.max_corner.lat) & lon_is_in_range



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Dispatch task on free workers
=============================
"""
from typing import Any, Callable, Iterator, List, Set
import time
import dask.distributed


def _available_workers(client: dask.distributed.Client) -> Set[str]:
    """Get the list of available workers.

    Args:
        client (dask.distributed.Client): Client connected to the Dask
        cluster.

    Returns:
        list: The list of available workers.
    """
    while True:
        info = client.scheduler_info()
        result = set(info['workers']) - set(
            k for k, v in client.processing().items() if v)
        if result:
            return result
        time.sleep(0.1)


def compute(client: dask.distributed.Client, func: Callable, seq: Iterator,
            *args, **kwargs) -> List[Any]:
    """Distribute the execution of functions to free workers, i.e. those who
    do not perform any tasks.

    Args:
        client (dask.distributed.Client): Client connected to the Dask
            cluster.
        func (callable): Function to execute
        seq (iterable): The sequence of arguments handled by ``func``.
        *args: Variable length argument list.
        **kwargs: Arbitrary keyword arguments.

    Returns:
        list: The result of the execution of the functions.
    """
    completed = dask.distributed.as_completed()
    workers = set()
    result = []
    iterate = True
    # As long as there is data to traverse in the iterator.
    while iterate:
        # As long as there are free workers
        while completed.count() < len(client.scheduler_info()['workers']):
            try:
                if not workers:
                    workers = _available_workers(client)
                item = next(seq)
                completed.add(
                    client.submit(func,
                                  item,
                                  *args,
                                  workers=workers.pop(),
                                  allow_other_workers=False,
                                  **kwargs))
            except StopIteration:
                iterate = False
                break
        # The computation queue is full, we consume the finished jobs to be
        # able to continue.
        if iterate:
            try:
                result += client.gather(completed.next_batch())
            except StopIteration:
                pass
    result += [item.result() for item in completed]
    return result



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Baseling dilation errors
------------------------
"""
from typing import Dict
import numpy as np

from .. import random_signal
from .. import settings
from .. import VOLUMETRIC_MEAN_RADIUS, BASELINE


class BaselineDilation:
    """
    Baseline dilation errors

    Args:
        parameters (settings.Parameters): Simulation settings
        dilation_psd (numpy.ndarray): Power spectral density of the baseline
            dilation.
        spatial_frequency (numpy.ndarray): Spatial frequency
    """
    def __init__(self, parameters: settings.Parameters,
                 dilation_psd: np.ndarray,
                 spatial_frequency: np.ndarray) -> None:
        # Store the generation parameters of the random signal.
        self.rng = parameters.rng()
        self.len_repeat = parameters.len_repeat
        self.delta_al = parameters.delta_al

        # Get baseline dilation power spectrum
        self.psbd = dilation_psd
        self.freq = spatial_frequency

        # TODO
        height = parameters.height * 1e-3
        self.conversion_factor = -((1 + height / VOLUMETRIC_MEAN_RADIUS) /
                                   (height * BASELINE)) * 1e-3

    def _generate_1d(self, x_al: np.ndarray) -> np.ndarray:
        # Generate 1d baseline dilation using the power spectrum:
        dil = random_signal.gen_signal_1d(self.freq,
                                          self.psbd,
                                          x_al,
                                          rng=self.rng,
                                          fmin=1 / self.len_repeat,
                                          fmax=1 / (2 * self.delta_al),
                                          alpha=10)

        # Compute the associated baseline dilation error on the swath in m
        return self.conversion_factor * dil

    def generate(self, x_al: np.ndarray,
                 x_ac: np.ndarray) -> Dict[str, np.ndarray]:
        """Generate the baseline dilation error

        Args:
            x_al (numpy.ndarray): Along track distance
            x_ac (numpy.ndarray): Across track distance

        Returns:
            dict: variable name and errors simulated.
        """
        baseline_dilation_1d = self._generate_1d(x_al)
        return {
            "simulated_error_baseline_dilation":
            x_ac**2 * baseline_dilation_1d[:, np.newaxis]
        }



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Generate instrumental errors
----------------------------
"""
from typing import Dict, Union
import dask.distributed
import numpy as np
from . import (Altimeter, BaselineDilation, CorrectedRollPhase, Karin, Orbital,
               RollPhase, Timing, WetTroposphere)
from .. import random_signal
from .. import settings


class Generator:
    """Instrumental error generator.

    Args:
        parameters (settings.Parameters): Simulation settings
        first_date (numpy.datetime64): Date of the first simulated
            measurement.
        orbit_duration (numpy.timedelta64, optional): Orbit duration.
    """
    def __init__(self, parameters: settings.Parameters,
                 first_date: np.datetime64, orbit_duration: np.timedelta64):
        #: The list of user-defined error generators
        self.generators = []

        assert parameters.error_spectrum is not None
        error_spectrum = random_signal.read_file_instr(
            parameters.error_spectrum, parameters.delta_al)

        for item in parameters.noise:
            if item == Altimeter.__name__:
                self.generators.append(Altimeter(parameters))
            elif item == BaselineDilation.__name__:
                self.generators.append(
                    BaselineDilation(parameters,
                                     error_spectrum['dilationPSD'].data,
                                     error_spectrum['spatial_frequency'].data))
            elif item == CorrectedRollPhase.__name__:
                self.generators.append(
                    CorrectedRollPhase(parameters, first_date))
            elif item == Karin.__name__:
                self.generators.append(Karin(parameters))
            elif item == Orbital.__name__:
                self.generators.append(Orbital(parameters, orbit_duration))
            elif item == RollPhase.__name__:
                self.generators.append(
                    RollPhase(parameters, error_spectrum['rollPSD'].data,
                              error_spectrum['gyroPSD'].data,
                              error_spectrum['phasePSD'].data,
                              error_spectrum['spatial_frequency'].data))
            elif item == Timing.__name__:
                self.generators.append(
                    Timing(parameters, error_spectrum['timingPSD'].data,
                           error_spectrum['spatial_frequency'].data))
            elif item == WetTroposphere.__name__:
                self.generators.append(WetTroposphere(parameters))
            else:
                # A new error generation class has been implemented but it is
                # not handled by this object.
                raise ValueError(f"unknown error generation class: {item}")

    def generate(self, cycle_number: int, pass_number: int,
                 curvilinear_distance: float, time: np.ndarray,
                 x_al: np.ndarray, x_ac: np.ndarray,
                 swh: Union[np.ndarray, float]) -> Dict[str, np.ndarray]:
        """Generate errors

        Args:
            cycle_number (int): Cycle number.
            pass_number (int): Pass number.
            curvilinear_distance (float): Curvilinear distance covered by the
                satellite during a complete cycle.
            time (numpy.ndarray): Date of measurements.
            x_al (numpy.ndarray): Along track distance.
            x_ac (numpy.ndarray): Across track distance.
            swh (numpy.ndarray, float): Significant wave height. Used to
                modulate instrumental noise as a function of sea state.

        Returns:
            dict: Associative array between error variables and simulated
            values.
        """
        # The variable "x_al" contains the distance a along track from the
        # beginning of the cycle. In order not to reproduce the same error from
        # one cycle to another, we transform it into the distance covered since
        # the first cycle.
        x_al += curvilinear_distance * (cycle_number - 1)

        result = {}
        if not self.generators or x_al.shape[0] == 0:
            return result

        futures = []
        with dask.distributed.worker_client() as client:
            for item in self.generators:
                if isinstance(item, Altimeter):
                    futures.append(client.submit(item.generate, x_al))
                elif isinstance(item, BaselineDilation):
                    futures.append(client.submit(item.generate, x_al, x_ac))
                elif isinstance(item, CorrectedRollPhase):
                    futures.append(client.submit(item.generate, time, x_ac))
                elif isinstance(item, Karin):
                    if isinstance(swh, float):
                        swh = np.full((x_ac.size, ), swh)
                    futures.append(
                        client.submit(item.generate,
                                      cycle_number * 10000 + pass_number, x_al,
                                      x_ac, swh))
                elif isinstance(item, Orbital):
                    futures.append(client.submit(item.generate, time, x_ac))
                elif isinstance(item, RollPhase):
                    futures.append(client.submit(item.generate, x_al, x_ac))
                elif isinstance(item, Timing):
                    futures.append(client.submit(item.generate, x_al, x_ac))
                elif isinstance(item, WetTroposphere):
                    futures.append(client.submit(item.generate, x_al, x_ac))

            for future in dask.distributed.as_completed(futures):
                result.update(future.result())
        return result



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Orbital error
-------------
"""
from typing import Dict, Tuple

#
import dask.array as da
import numpy as np

#
from .. import random_signal
from .. import settings
from .. import VOLUMETRIC_MEAN_RADIUS

#: Signal amplitude of the orbital error in micro-radians
AMPLITUDE = 100

#: Delta T of the spatial sampling in seconds
DT = 60


def _orbital_error_spectrum(
        orbit_duration: np.timedelta64,
        rng: np.random.Generator) -> Tuple[np.ndarray, float]:
    """Calculate orbital error spectrum

    Args:
        orbit_duration (float): Orbit duration in fractional days
        rng (np.random.Generator): Random number generator

    Returns:
        tuple: (yg, fmaxr)
    """
    df = 1 / (1000 * 86400)
    spatial_frequency = np.arange(df, 1 / DT, df)
    orbital_frequency = 1 / float(
        orbit_duration.astype("timedelta64[us]").astype("float64") * 1e-6)
    sigma_peak = orbital_frequency / 1000
    ps_orbital = np.exp(-0.5 * (spatial_frequency - orbital_frequency)**2 /
                        sigma_peak**2)
    ps_orbital[ps_orbital < 1 / 1000] = 0.
    ps_orbital /= np.sum(ps_orbital * df)
    ps_orbital *= AMPLITUDE**2
    return random_signal.gen_psd_1d(spatial_frequency,
                                    ps_orbital,
                                    rng,
                                    alpha=10)


class Orbital:
    """
    Simulate the error orbital

    Args:
        parameters (Parameters): Simulation parameters.
        orbit_duration (np.timedelta64): Orbit duration.
    """
    def __init__(self, parameters: settings.Parameters,
                 orbit_duration: np.timedelta64) -> None:
        yg, self.fmaxr = _orbital_error_spectrum(orbit_duration,
                                                 parameters.rng())
        self.yg = da.from_array(yg, name="orbital_error").persist()

        assert parameters.height is not None
        height = parameters.height * 1e-3
        self.conversion_factor = (1 + height / VOLUMETRIC_MEAN_RADIUS) * 1e-3

    def generate(
        self,
        time: np.ndarray,
        x_ac: np.ndarray,
    ) -> Dict[str, np.ndarray]:
        """Generate orbital error

        Args:
            time (np.ndarray): time vector

        Returns:
            np.ndarray: orbital error
        """
        time = time.astype("datetime64[us]").astype("float64") * 1e-6
        xg = np.linspace(0, 0.5 / self.fmaxr * self.yg.shape[0],
                         self.yg.shape[0])
        error_orbital = np.interp(np.mod(time, xg.max()), xg,
                                  self.yg.compute())
        return {
            "simulated_error_orbital":
            x_ac * error_orbital[:, np.newaxis] * self.conversion_factor,
        }



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Main program
------------

This module defines the main :func:`function <launch>` handling the simulation
of SWOT products as well as the entry point of the main program.
"""
from typing import Dict, List, Optional, Tuple
import argparse
import datetime
import importlib
import logging
import os
import pathlib
import platform
import sys
import traceback

import dask.distributed
import dateutil.parser
import numpy as np

from . import (dispatch, exception, logbook, netcdf, orbit_propagator,
               settings, __date__, __version__)
from .error import generator

#: Logger of this module
LOGGER = logging.getLogger(__name__)

#: Tasks launched
TASKS = {}


def datetime_type(value: str) -> np.datetime64:
    """The option should define a datetime

    Args:
        value (str): Value to parse

    Returns:
        numpy.datetime64: The date parsed

    Raises:
        ValueError: if the option provided does not set a valid date.
    """
    # Check if value is a datetime
    try:
        result = dateutil.parser.parse(value)
    except ValueError as error:
        raise argparse.ArgumentTypeError(
            f"invalid date time {value!r}: {error!s}")
    return np.datetime64(result)


def usage() -> argparse.Namespace:
    """Parse the options provided on the command line.

    Returns:
        argparse.Namespace: The parameters provided on the command line.
    """
    parser = argparse.ArgumentParser(add_help=False)
    parser.add_argument('-h',
                        '--help',
                        action='store_true',
                        help='show this help message and exit')
    group = parser.add_argument_group("General", "Simulation general settings")
    group.add_argument("--first-date",
                       help="The first date to be processed. "
                       "Default to the current date",
                       type=datetime_type,
                       default=np.datetime64("now"))
    group.add_argument("--last-date",
                       help="The last date to be processed. "
                       "Default to the last date allowing to cover an entire "
                       "cycle.",
                       type=datetime_type)
    group = parser.add_argument_group("Execution",
                                      "Runtime parameters options ")
    group.add_argument("--debug",
                       action="store_true",
                       help="Put swot simulator in debug mode")
    group.add_argument("--log",
                       metavar='PATH',
                       help="Path to the logbook to use",
                       type=argparse.FileType("w"))
    group.add_argument("--scheduler-file",
                       help="Path to a file with scheduler information to "
                       "launch swot simulator on a cluster. By "
                       "default, use a local cluster.",
                       metavar='PATH',
                       type=argparse.FileType("r"))
    group = parser.add_argument_group("LocalCluster",
                                      "Dask local cluster option")
    group.add_argument("--n-workers",
                       help="Number of workers to start (Default to 1)",
                       type=int,
                       metavar='N',
                       default=1)
    group.add_argument("--processes",
                       help="Whether to use processes (True) or threads "
                       "(False).  Defaults to False",
                       action="store_true")
    group.add_argument("--threads-per-worker",
                       help="Number of threads per each worker. "
                       "(Default to 1)",
                       type=int,
                       metavar='N',
                       default=1)
    group = parser.add_argument_group("Configuration")
    group.add_argument("--template",
                       help="Writes the default configuration of the "
                       "simulator into the file and ends the program.",
                       metavar="PATH",
                       type=argparse.FileType("w"))
    namespace = argparse.Namespace()
    namespace, _ = parser._parse_known_args(sys.argv[1:], namespace)

    def add_settings(parser):
        """Added the argument defining the settings of the simulator."""
        parser.add_argument("settings",
                            type=argparse.FileType('r'),
                            help="Path to the parameters file")

    # Displays help and ends the program.
    if "help" in namespace:
        add_settings(parser)
        parser.print_help()
        parser.exit(0)

    # Checking exclusive options.
    if "scheduler_file" in namespace:
        for item in ["n_workers", "processes", "threads_per_worker"]:
            if item in namespace:
                item = item.replace("_", "-")
                raise RuntimeError(
                    f"--{item}: not allowed with argument --scheduler-file")

    # Write the template configurayion file and ends the programm
    if "template" in namespace:
        namespace.template.write(settings.template())
        sys.stdout.write(f"""
The template has been written in the file: {namespace.template.name!r}.
""")
        parser.exit(0)

    # The partial analysis of the command line arguments is finished, the last
    # argument is added and parsed one last time.
    add_settings(parser)

    return parser.parse_args()


def file_path(first_date: np.datetime64,
              last_date: np.datetime64,
              cycle_number: int,
              pass_number: int,
              parameters: settings.Parameters,
              nadir: bool = False) -> Optional[str]:
    """Get the absolute path of the file to be created.

    Args:
        first_date (numpy.datetime64): Date of the first simulated measurement.
        last_date (numpy.datetime64): Date of the last simulated measurement.
        cycle_number (int): Cycle number of the file to be created.
        pass_number (int): Pass number of the file to be created.
        parameters (settings.Parameters): Simulation parameters.
        nadir (bool, optional): True if the product to be created contains the
            nadir measurements, false if it is a product containing the swath
            measurements.

    Returns:
        str, optional: The path to the file to be created or None if the file
        already exists
    """
    first_datetime: datetime.datetime = first_date.astype(
        datetime.datetime)  # type: ignore
    last_datetime: datetime.datetime = last_date.astype(
        datetime.datetime)  # type: ignore
    product_type = "nadir" if nadir else "karin"
    dirname = os.path.join(parameters.working_directory, product_type,
                           first_datetime.strftime("%Y"))
    os.makedirs(dirname, exist_ok=True)
    if nadir:
        filename = (f"SWOT_GPR_2P1P{cycle_number:03d}_{pass_number:03d}_"
                    f"{first_datetime:%Y%m%d}_{first_datetime:%H%M%S}_"
                    f"{last_datetime:%Y%m%d}_{last_datetime:%H%M%S}.nc")
    else:
        product_type = "".join(
            [item.capitalize() for item in parameters.product_type.split("_")])
        filename = (f"SWOT_L2_LR_SSH_{product_type}_"
                    f"{cycle_number:03d}_{pass_number:03d}_"
                    f"{first_datetime:%Y%m%dT%H%M%S}_"
                    f"{last_datetime:%Y%m%dT%H%M%S}_"
                    "DG10_01.nc")
    result = os.path.join(dirname, filename)
    # If the product has already been produced, the generation of this
    # half orbit is disabled.
    if os.path.exists(result):
        LOGGER.debug("%s already generated: %d/%d",
                     "nadir" if nadir else "swath", cycle_number, pass_number)
        return None
    return result


def sum_error(errors: Dict[str, np.ndarray], swath: bool = True) -> np.ndarray:
    """Calculate the sum of errors

    Args:
        errors (dict): Simulated errors.
        swath (bool, optional): True if the measurements processed concern the
            swath.

    Returns:
        numpy.ndarray: The sum of errors
    """
    dims = 2 if swath else 1
    return np.add.reduce(
        [item for item in errors.values() if len(item.shape) == dims])


def simulate(args: Tuple[int, int, np.datetime64],
             error_generator: generator.Generator,
             orbit: orbit_propagator.Orbit,
             parameters: settings.Parameters,
             logging_server: Optional[Tuple[str, int, int]] = None) -> None:
    """Simulate a pass.

    Args:
        simulation arguments (tuple): cycle & pass number, date of the first
            half-orbit measurement
        error_generator (generator.Generator): Measurement error generator.
        orbit (orbit_propagator.Orbit): Orbit propagator.
        parameters (settings.Parameters): Simulation parameters.
        logging_server (tuple, optional): Log server connection settings.
    """
    cycle_number, pass_number, date = args

    # Initialize this worker's logger.
    if logging_server is not None:
        logbook.setup_worker_logging(logging_server)

    # Calculation of the end date of the track.
    last_date = date + orbit.pass_shift(pass_number)

    # Paths of products to be generated.
    swath_path = None
    nadir_path = None

    # Generation of the names of the files to be created.
    if parameters.swath:
        swath_path = file_path(date, last_date, cycle_number, pass_number,
                               parameters)

    if parameters.nadir:
        nadir_path = file_path(date,
                               last_date,
                               cycle_number,
                               pass_number,
                               parameters,
                               nadir=True)

    # To continue, there must be at least one task left for this pass.
    if swath_path is None and nadir_path is None:
        return

    # Compute the spatial/temporal position of the satellite
    track = orbit_propagator.calculate_pass(pass_number, orbit, parameters)
    if track is None:
        return

    # Set the simulated date
    track.set_simulated_date(date)

    # Mask to set the measurements outside the requirements of the mission to
    # NaN.
    mask = track.mask()

    # The nadir and swath data are concatenated to process the interpolation
    # in one time (swath and nadir).
    lon = np.c_[track.lon, track.lon_nadir[:, np.newaxis]]
    lat = np.c_[track.lat, track.lat_nadir[:, np.newaxis]]
    swath_time = np.repeat(track.time, lon.shape[1]).reshape(lon.shape)

    # Interpolation of the SWH if the user wishes.
    if parameters.swh_plugin is not None:
        swh = parameters.swh_plugin.interpolate(lon.flatten(), lat.flatten(),
                                                swath_time.flatten())
        swh = swh.reshape(lon.shape)
    else:
        swh = None

    # Calculation of instrumental errors
    noise_errors = error_generator.generate(
        cycle_number, pass_number, orbit.curvilinear_distance, track.time,
        track.x_al, track.x_ac,
        swh[:, :-1] if swh is not None else parameters.swh)
    for error in noise_errors.values():
        # Only the swaths must be masked
        if len(error.shape) == 2:
            error *= mask

    # Interpolation of the SSH if the user wishes.
    if parameters.ssh_plugin is not None:
        ssh = parameters.ssh_plugin.interpolate(lon.flatten(), lat.flatten(),
                                                swath_time.flatten())
        ssh = ssh.reshape(lon.shape)
    else:
        ssh = None

    if swath_path:
        LOGGER.info("generate swath %d/%d [%s, %s]", cycle_number, pass_number,
                    track.time[0], track.time[-1])

        # Create the swath dataset
        product = netcdf.Swath(track, parameters.central_pixel,
                               parameters.product_type)

        if ssh is not None:
            product.ssh((ssh[:, :-1] * mask) + sum_error(noise_errors))
            if noise_errors:
                product.simulated_true_ssh(ssh[:, :-1])
        if swh is not None:
            product.swh((swh[:, :-1] * mask))

        product.update_noise_errors(noise_errors)
        product.to_netcdf(cycle_number, pass_number, swath_path,
                          parameters.complete_product)

    # Create the nadir dataset
    if nadir_path:
        LOGGER.info("generate nadir %d/%d [%s, %s]", cycle_number, pass_number,
                    track.time[0], track.time[-1])

        product = netcdf.Nadir(track)

        if ssh is not None:
            product.ssh(ssh[:, -1] + sum_error(noise_errors, swath=False))
            if noise_errors:
                product.simulated_true_ssh(ssh[:, -1])
        if swh is not None:
            product.swh(swh[:, -1])
        product.update_noise_errors(noise_errors)
        product.to_netcdf(cycle_number, pass_number, nadir_path,
                          parameters.complete_product)


def launch(client: dask.distributed.Client,
           parameters: settings.Parameters,
           logging_server: Optional[Tuple[str, int, int]],
           first_date: Optional[np.datetime64] = None,
           last_date: Optional[np.datetime64] = None):
    """Executes the simulation set to the selected period.

    Args:
        client (dask.distributed.Client): Client connected to the Dask
            cluster.
        parameters (settings.Parameters): Simulation parameters.
        logging_server (tuple, optional): Log server connection settings.
        first_date (numpy.datetime64, optional): First date of the simulation.
        last_date (numpy.datetime64, optional): Last date of the simulation.
    """
    first_date = first_date or np.datetime64("now")

    # Displaying Dask client information.
    LOGGER.info(client)

    assert parameters.ephemeris is not None

    # Calculation of the properties of the orbit to be processed.
    with open(parameters.ephemeris, "r") as stream:  # type: TextIO
        orbit = orbit_propagator.calculate_orbit(parameters,
                                                 stream)  # type: ignore

    # Initialization of measurement error generators
    error_generator = generator.Generator(parameters, first_date,
                                          orbit.orbit_duration())

    # Scatter data into distributed memory
    _error_generator = client.scatter(error_generator)
    _parameters = client.scatter(parameters)
    _orbit = client.scatter(orbit)

    dispatch.compute(client,
                     simulate,
                     orbit.iterate(first_date, last_date),
                     error_generator=_error_generator,
                     orbit=_orbit,
                     parameters=_parameters,
                     logging_server=logging_server)


def software_dependencies() -> List[Tuple[str, str]]:
    """Returns the software dependencies of this release"""
    deps = [
        ("conda", lambda moduleule: moduleule.__version__),
        ("dask", lambda module: module.__version__),
        ("distributed", lambda module: module.__version__),
        ("netCDF4", lambda module: module.__version__),
        ("numba", lambda module: module.__version__),
        ("numpy", lambda module: module.__version__),
        ("pip", lambda module: module.__version__),
        ("pyinterp", lambda module: module.__version__),
        ("scipy", lambda module: module.__version__),
        ("setuptools", lambda module: module.__version__),
        ("xarray", lambda module: module.__version__),
        ("swot_simulator", lambda module: module.__version__),
    ]
    result: List[Tuple[str, str]] = [("python", platform.python_version())]
    for module_name, version_getter in deps:
        try:
            if module_name in sys.modules:
                mod = sys.modules[module_name]
            else:
                mod = importlib.import_module(module_name)
            version = version_getter(mod)
            result.append((module_name, version))
        except ImportError:
            pass
    return result


def copy_parameters(parameters: settings.Parameters,
                    path: pathlib.Path) -> None:
    """Copies the overriden parameters into the working directory.

    Args:
        parameters (settings.Parameters): Simulation parameters.
        filename (pathlib.Path): Path to the overriden parameters file used.
    """
    target = pathlib.Path(parameters.working_directory).joinpath(path.name)
    target.parent.mkdir(parents=True, exist_ok=True)
    version = 1
    now = datetime.datetime.now().isoformat()
    while target.exists():
        target = pathlib.Path(str(target.with_suffix("")) + f" ({version}).py")
        version += 1

    with target.open("w") as target_stream:
        with path.open("r") as source_stream:
            target_stream.write(f"""
# This file has been automatically generated by the SWOT simulator on {now}.
# It contains the overriden parameters used to run the simulation.
# Please do not modify it manually.
#
# Dependency versions
# -------------------
""")
            for dependency, version in sorted(software_dependencies()):
                target_stream.write(f"# * {dependency}: {version}\n")
            target_stream.write(source_stream.read())


def main():
    """main function"""
    args = usage()

    # Setup log
    logger, logging_server = logbook.setup(args.log, args.debug)

    client = dask.distributed.Client(
        dask.distributed.LocalCluster(
            protocol="tcp",
            n_workers=args.n_workers,
            processes=args.processes,
            threads_per_worker=args.threads_per_worker)
    ) if args.scheduler_file is None else dask.distributed.Client(
        scheduler_file=args.scheduler_file.name)

    client.wait_for_workers(1)

    try:
        parameters = settings.Parameters(
            settings.eval_config_file(args.settings.name))

        # Keep track of the overriden parameters
        copy_parameters(parameters, pathlib.Path(args.settings.name))
        launch(client, parameters, logging_server, args.first_date,
               args.last_date)

        client.close()
        logger.info("End of processing.")
        return 0
    #: pylint: disable=broad-except
    # All exceptions are caught in the main function to display it in the log
    # and return the appropriate error code.
    except Exception as exc:
        # Clients are stopped before writing the error message to the log to
        # ensure that the log will end with the exception captured.
        client.close()
        logger.error(
            exception.structured_traceback(
                exc, traceback.extract_tb(sys.exc_info()[2])))
        logger.error("End of processing.")
    #: pylint: enable=broad-except
    return 1


if __name__ == '__main__':
    sys.exit(main())



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Wet troposphere errors
----------------------
"""
from typing import Dict, List, Tuple
import numba as nb
import numba.typed
import numpy as np
import scipy.ndimage.filters

from .. import random_signal
from .. import settings
from .. import F_KA, VOLUMETRIC_MEAN_RADIUS, CELERITY, BASELINE


@nb.njit(cache=True, nogil=True)
def _meshgrid(x: np.ndarray, y: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    nx = x.size
    ny = y.size
    mx = np.empty((ny, nx))
    my = np.empty((ny, nx))
    for idx in range(ny):
        mx[idx, :] = x
    for idx in range(nx):
        my[:, idx] = y
    return mx, my


@nb.njit(cache=True, nogil=True)
def _calculate_path_delay_lr(beam_positions: List[float], sigma: float,
                             radio_r: np.ndarray, radio_l: np.ndarray,
                             x_al: np.ndarray, x_ac_large: np.ndarray,
                             wt_large: np.ndarray):
    beam_r = np.empty((x_al.shape[0], ))
    beam_l = np.empty((x_al.shape[0], ))
    # Find righ and leftacross track indices in the gaussian
    # footprint of 2.*p.sigma
    ind_r = x_ac_large + beam_positions[1]
    indac_r = np.where((ind_r < 2 * sigma) & (ind_r > -2 * sigma))[0]
    ind_l = x_ac_large + beam_positions[0]
    indac_l = np.where((ind_l < 2 * sigma) & (ind_l > -2 * sigma))[0]

    factor = 1 / (2 * np.pi * sigma**2)

    for idx, xal in enumerate(x_al):
        # Find along track indices in the gaussian footprint
        # of 2.*p.sigma
        delta_x_al = x_al - xal
        indal = np.where((delta_x_al <= (2 * sigma))
                         & (delta_x_al > (-2 * sigma)))[0]
        slice_al = slice(indal[0], indal[-1] + 1)
        slice_acr = slice(indac_r[0], indac_r[-1] + 1)
        slice_acl = slice(indac_l[0], indac_l[-1] + 1)
        x, y = _meshgrid(x_ac_large[slice_acr], x_al[slice_al] - xal)
        # Compute path delay on left and right gaussian footprint
        g = factor * np.exp(-(x**2 + y**2) / (2 * sigma**2))
        beam_r[idx] = np.sum(
            g * wt_large[slice_al, slice_acr]) / np.sum(g) + radio_r[idx]
        x, y = _meshgrid(x_ac_large[slice_acl], x_al[slice_al] - xal)
        g = factor * np.exp(-(x**2 + y**2) / (2 * sigma**2))
        beam_l[idx] = np.sum(
            g * wt_large[slice_al, slice_acl]) / np.sum(g) + radio_l[idx]
    return beam_r, beam_l


@nb.njit(cache=True, nogil=True)
def _calculate_path_delay(sigma: float, radio: np.ndarray, x_al: np.ndarray,
                          x_ac_large: np.ndarray, wt_large: np.ndarray):
    beam = np.empty((x_al.shape[0], ))
    # Find across track indices in the gaussian footprint of
    # 2. * sigma
    indac = np.where((x_ac_large < 2 * sigma) & (x_ac_large > -2 * sigma))[0]
    factor = 1 / (2 * np.pi * sigma**2)
    for idx, xal in enumerate(x_al):
        delta_x_al = x_al[:] - xal
        indal = np.where((delta_x_al <= (2 * sigma))
                         & (delta_x_al > (-2 * sigma)))[0]
        slice_al = slice(indal[0], indal[-1] + 1)
        slice_ac = slice(indac[0], indac[-1] + 1)
        x, y = _meshgrid(x_ac_large[slice_ac], x_al[slice_al] - xal)
        # Compute path delay on gaussian footprint
        g = factor * np.exp(-(x**2 + y**2) / (2 * sigma**2))
        beam[idx] = np.sum(
            g * wt_large[slice_al, slice_ac]) / np.sum(g) + radio[idx]
    return beam


class WetTroposphere:
    """Wet troposphere errors

    Args:
        parameters (settings.Parameters): Simulation settings
    """
    ALPHA = 10
    LC_MAX = 500
    F_MAX = 0.05

    def __init__(self, parameters: settings.Parameters) -> None:
        # Store the generation parameters of the random signal.
        self.beam_positions = parameters.beam_position
        self.delta_ac = parameters.delta_ac
        self.delta_al = parameters.delta_al
        self.len_repeat = parameters.len_repeat
        self.nbeam = parameters.nbeam
        self.rng = parameters.rng()
        self.rng_radio_l = parameters.rng()
        self.rng_radio_r = parameters.rng()
        self.sigma = parameters.sigma
        # TODO
        self.conversion_factor = (
            1 / (F_KA * 2 * np.pi / CELERITY * BASELINE) *
            (1 + (parameters.height * 1e-3) / VOLUMETRIC_MEAN_RADIUS) * np.pi /
            180 * 1e3)
        # Define power spectrum of error in path delay due to wet tropo
        freq = np.arange(1 / 3000, 1 / (2 * self.delta_al), 1 / 3000)
        # Global mean wet tropo power spectrum in cm**2/(cycle/km)
        # for L >= 100 km
        pswt = 3.156 * 1e-05 * freq**(-8 / 3)
        # Wet tropo power spectrum in cm**2/(cycle/km) for L < 100 km
        mask = freq > 1e-2
        pswt[mask] = 1.4875 * 1e-4 * freq[mask]**(-2.33)
        self.pswt = pswt
        self.freq = freq
        self.fminx = 1 / self.len_repeat
        self.ps2d, self.f = random_signal.gen_ps2d(freq,
                                                   pswt,
                                                   fminx=self.fminx,
                                                   fminy=1 / self.LC_MAX,
                                                   fmax=self.F_MAX,
                                                   alpha=self.ALPHA,
                                                   lf_extpl=True,
                                                   hf_extpl=True)

    def _radiometer_error(self,
                          x_al: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
        # Define radiometer error power spectrum for a beam
        # High frequencies are cut to filter the associated error:
        # during the reconstruction of the wet trop signal
        psradio = 9.5 * 1e-5 * self.freq**-1.79
        psradio[self.freq < 1e-3] = 9.5 * 1e-5 * (1e-3)**-1.79
        mask = (self.freq > 0.0023) & (self.freq <= 0.0683)
        psradio[mask] = 0.036 * self.freq[mask]**-0.814
        psradio[self.freq > 0.0683] = 0.32
        # Compute random coefficients (1D) for the radiometer error
        # power spectrum for right and left beams
        hrad = random_signal.gen_signal_1d(self.freq,
                                           psradio,
                                           x_al,
                                           fmin=1 / self.len_repeat,
                                           fmax=1 / (2 * self.delta_al),
                                           alpha=10,
                                           rng=self.rng_radio_r,
                                           hf_extpl=True,
                                           lf_extpl=True)
        radio_r = hrad * 1e-2
        hrad = random_signal.gen_signal_1d(self.freq,
                                           psradio,
                                           x_al,
                                           fmin=1 / self.len_repeat,
                                           fmax=1 / (2 * self.delta_al),
                                           alpha=10,
                                           rng=self.rng_radio_l,
                                           hf_extpl=True,
                                           lf_extpl=True)
        radio_l = hrad * 1e-2

        return radio_r, radio_l

    def generate(self, x_al: np.array,
                 x_ac: np.array) -> Dict[str, np.ndarray]:
        """
        Generate wet troposphere errors

        Args:
            x_al (numpy.ndarray): Along track distance
            x_ac (numpy.ndarray): Across track distance

        Returns:
            dict: variable name and errors simulated.
        """
        num_lines = x_al.shape[0]
        num_pixels = x_ac.shape[0]

        # Initialization of radiometer error in right and left beam
        radio_r, radio_l = self._radiometer_error(x_al)
        # Initialization of swath matrices and large swath matrices (which
        # include wet tropo data around the nadir and outside the swath)
        start_x = -2 * self.sigma / self.delta_ac + x_ac[0]
        stop_x = (2 * self.sigma / self.delta_ac + x_ac[-1] + self.delta_ac)
        # x_ac_large and wt_large are necessary to compute the gaussian
        # footprint of a beam on the nadir or near the edge of the swath
        x_ac_large = np.arange(start_x, stop_x, self.delta_ac)
        naclarge = np.shape(x_ac_large)[0]
        # Compute path delay error due to wet tropo and radiometer error
        # using random coefficient initialized with power spectrums
        wt = random_signal.gen_signal_2d_rectangle(self.ps2d,
                                                   self.f,
                                                   x_al,
                                                   x_ac,
                                                   fminx=self.fminx,
                                                   fminy=1 / self.LC_MAX,
                                                   fmax=self.F_MAX,
                                                   alpha=self.ALPHA,
                                                   rng=self.rng)
        wt = wt.T * 1e-2
        wt_large = random_signal.gen_signal_2d_rectangle(self.ps2d,
                                                         self.f,
                                                         x_al,
                                                         x_ac_large,
                                                         fminx=self.fminx,
                                                         fminy=1 / self.LC_MAX,
                                                         fmax=self.F_MAX,
                                                         alpha=self.ALPHA,
                                                         rng=self.rng)
        wt_large = wt_large.T * 1e-2

        # Compute Residual path delay error after a 1-beam radiometer
        # correction
        if self.nbeam == 1:
            beam = _calculate_path_delay(self.sigma, radio_l, x_al, x_ac_large,
                                         wt_large)
            beam = scipy.ndimage.filters.gaussian_filter(
                beam, 30. / self.delta_al)
            beam2d = np.vstack(num_pixels * (beam, )).T
            # Compute residual path delay
            wet_tropo = wt - beam2d
            wet_tropo_nadir = wt_large[:, naclarge // 2] - beam

        # Compute Residual path delay error after a 2-beams radiometer
        # correction
        elif self.nbeam == 2:
            beam_r, beam_l = _calculate_path_delay_lr(
                numba.typed.List(self.beam_positions), self.sigma, radio_r,
                radio_l, x_al, x_ac_large, wt_large)
            # Filtering beam signal to cut frequencies higher than 125 km
            beam_r = scipy.ndimage.filters.gaussian_filter(
                beam_r, 30 / self.delta_al)
            beam_l = scipy.ndimage.filters.gaussian_filter(
                beam_l, 30 / self.delta_al)
            # Compute residual path delay (linear combination of left
            # and right path delay)
            polyfit = np.polynomial.polynomial.polyfit
            pol = polyfit([self.beam_positions[0], self.beam_positions[1]],
                          [beam_l, beam_r], 1)
            beam = (np.array(num_pixels * [pol[0]]).T +
                    np.array(num_lines * [x_ac]) *
                    np.array(num_pixels * [pol[1]]).T)
            wet_tropo = wt - beam
            wet_tropo_nadir = wt_large[:, naclarge //
                                       2] - beam[:, num_pixels // 2]
        else:
            raise ValueError("nbeam must be in [1, 2]")

        # wt_nadir = wt_large[:, naclarge // 2]

        return {
            "simulated_error_troposphere": wet_tropo,
            "simulated_error_troposphere_nadir": wet_tropo_nadir
        }



# Configuration file used to generate the SWOT_L2_LR_SSH_Expert files according
# to the SWOT PDD.
import pathlib
import swot_simulator

# Geographical area to simulate defined by the minimum and maximum corner
# point :lon_min, lat_min, lon_max, lat_max. Default: -180, -90, 180, 90
area = None

# Number of beam used to correct wet troposphere signal (1, 2 or 'both')
beam_position = [-20, 20]

# If true, the swath, in the final dataset, will contain a center pixel
# divided in half by the reference ground track
central_pixel = True

# If true, the generated netCDF file will be the complete product compliant
# with SWOT's Product Description Document (PDD), otherwise only the calculated
# variables will be written to the netCDF file
complete_product = True

# Duration of a cycle in number of fractional days. By default, this value
# is read from the ephemeris file)
# cycle_duration = None

# Distance, in km, between two points across track direction
delta_ac = 2.0

# Distance, in km, between two points along track direction
delta_al = 2.0

# Index of columns to read in the ephemeris file containing, respectively,
# longitude in degrees, latitude in degrees and the number of seconds elapsed
# since the start of the orbit. Default: [1, 2, 0]
ephemeris_cols = None

# Ephemeris file to read containing the satellite's orbit.
#ephemeris = os.path.join(DATA, 'ephemeris_calval_june2015_ell.txt')
ephemeris = swot_simulator.DATA / 'ephemeris_science_sept2015_ell.txt'

# File containing spectrum of instrument error
error_spectrum = swot_simulator.DATA / 'error_spectrum.nc'

# Estimated roll phase dataset. Default: None
corrected_roll_phase_dataset = None

# Distance, in km, between the nadir and the center of the first pixel
# of the swath
half_gap = 2.0

# Distance, in km, between the nadir and the center of the last pixel
# of the swath
half_swath = 70.0

# Satellite altitude (m). By default, this value is read from the ephemeris
# file.
# height = None

# KaRIN file containing spectrum for several SWH
karin_noise = swot_simulator.DATA / 'karin_noise_v2.nc'

# Repeat length
len_repeat = 20000.0

# True to generate Nadir products
nadir = True

# Number of beam used to correct wet troposphere signal (1, 2 or 'both')
nbeam = 2

# This option defined the error to simulate. Allowed values are "altimeter,"
# "baseline_dilation," "corrected_roll_phase," "karin," "orbital," "roll_phase,"
# "timing," and "wet_troposphere." The calculation of roll errors can be
# simulated, option "roll_phase," or interpolated option "corrected_roll_phase,"
# from the dataset specified by the value of the option "roll_phase_dataset."
# Therefore, these two options are mutually exclusive. In other words, if
# the "roll_phase" option is present, the "corrected_roll_phase" option must
# be omitted, and vice versa.
noise = [
    'altimeter',
    'baseline_dilation',
    'karin',
    'orbital',
    'roll_phase',
    'timing',
    # 'wet_troposphere',
]

# Seed for RandomState. Must be convertible to 32 bit unsigned integers
nseed = 0

# Type of SWOT product to be generated. Possible products are 'basic', 'expert',
# 'unsmoothed' and 'wind_wave'. Default to 'expert'
product_type = 'expert'

# Limits of SWOT swath requirements. Measurements outside the span will be
# set with fill values
requirement_bounds = [10, 60]

# Orbit shift in longitude (degrees). Default 0.
shift_lon = None

# Orbit shift in time (seconds). Default 0.
shift_time = None

# Gaussian footprint of sigma (km)
sigma = 6.0

# The plug-in handling the SSH interpolation under the satellite swath
ssh_plugin = None

# SWH plugin to interpolate model SWH on the SWOT grid. Use only 'expert'
# or 'wind_wave' products that have the swh in its output
swh_plugin = None

# True to generate swath products
swath = True

# SWH for the region
swh = 2.0

# The working directory. By default, files are generated in the user's root
# directory
working_directory = None



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Exception handlers
------------------
"""
import linecache
import os
import traceback


def structured_traceback(exc: Exception,
                         call_stack: traceback.StackSummary) -> str:
    """Return a nice text describing the call stack which threw an
    exception.

    Args:
        exc (Exception): Exception raised.
        call_stack (traceback.StackSummary): Exception call stack thrown out.

    Returns:
        str: The text representing the call stack of the thrown exception.
    """
    exc_name = "%s.%s" % (exc.__module__, exc.__class__.__name__) if hasattr(
        exc, '__module__') else exc.__class__.__name__
    message = ["%s - Traceback (most recent call last):" % exc_name]

    for source, line_number, function, _ in call_stack:
        message.append("%s in %s" % (os.path.abspath(source), function))

        # Reading the Python code
        lines = linecache.getlines(source)

        # We will display the Python source code around the call so that
        # the user visualizes the incriminated lines well.
        start = line_number - 3
        end = line_number + 2
        start = max(start, 0)

        # Add the Python source code
        for idx, item in enumerate(lines[start:end]):
            where = start + idx + 1
            message.append(("----> " if where == line_number else "      ") +
                           str(where) + " " + item.rstrip())
        message.append("")

    # Finally, displays the error message
    message.append(exc_name + ": " + str(exc))
    return "\n".join(message)



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Timing errors
-------------
"""
from typing import Dict
import numpy as np
from .. import random_signal

from .. import CELERITY
from .. import settings


class Timing:
    """
    Timing errors

    Args:
        parameters (settings.Parameters): Simulation settings
        timing_psd (numpy.ndarray): Power spectral density the timing error
        spatial_frequency (numpy.ndarray): Spatial frequency
    """
    CONVERSION_FACTOR = CELERITY * 5e-13

    def __init__(self, parameters: settings.Parameters, timing_psd: np.ndarray,
                 spatial_frequency: np.ndarray) -> None:
        # Store the generation parameters of the random signal.
        self.rng_l = parameters.rng()
        self.rng_r = parameters.rng()
        self.len_repeat = parameters.len_repeat
        self.delta_al = parameters.delta_al

        # Get baseline dilation power spectrum
        self.timing_psd = timing_psd
        self.spatial_frequency = spatial_frequency

    def _generate_1d(self, x_al: np.ndarray) -> np.ndarray:
        # Generate 1d timing using the power spectrum:
        timing_l = random_signal.gen_signal_1d(self.spatial_frequency,
                                               self.timing_psd,
                                               x_al,
                                               rng=self.rng_l,
                                               fmin=1 / self.len_repeat,
                                               fmax=1 / (2 * self.delta_al),
                                               alpha=10)
        timing_r = random_signal.gen_signal_1d(self.spatial_frequency,
                                               self.timing_psd,
                                               x_al,
                                               rng=self.rng_r,
                                               fmin=1 / self.len_repeat,
                                               fmax=1 / (2 * self.delta_al),
                                               alpha=10)
        # Compute the corresponding timing error on the swath in m
        return np.array([
            self.CONVERSION_FACTOR * timing_l,
            self.CONVERSION_FACTOR * timing_r
        ]).T

    def generate(self, x_al: np.ndarray,
                 x_ac: np.ndarray) -> Dict[str, np.ndarray]:
        """Generate timing errors

        Args:
            x_al (numpy.ndarray): Along track distance
            x_ac (numpy.ndarray): Across track distance

        Returns:
            dict: variable name and errors simulated.
        """
        timing_1d = self._generate_1d(x_al)
        num_pixels = x_ac.shape[0]
        swath_center = num_pixels // 2
        num_lines = timing_1d.shape[0]
        timing = np.empty((num_lines, num_pixels))
        ones_ac = np.ones((swath_center, ))
        timing[:, :swath_center] = ones_ac * timing_1d[:, 0, np.newaxis]
        timing[:, swath_center:] = ones_ac * timing_1d[:, 1, np.newaxis]

        return {"simulated_error_timing": timing}



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Simulate errors
---------------
"""
from .altimeter import Altimeter
from .baseline_dilation import BaselineDilation
from .corrected_roll_phase import CorrectedRollPhase
from .karin import Karin
from .orbital import Orbital
from .roll_phase import RollPhase
from .timing import Timing
from .wet_troposphere import WetTroposphere



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Interpolation of the SSH SYMPHONIE
==================================
"""
from .. import data_handler


class SYMPHONIE(data_handler.IrregularGridHandler):
    """
    Interpolation of the SSH SYMPHONIE
    """
    def __init__(self, path: str):
        loader = data_handler.NetcdfLoader(
            path,
            date_fmt="%Y%m%d_%Hh",
            lon_name="lon_t",
            lat_name="lat_t",
            ssh_name="ssh_ib_detided",
            pattern=r"(?P<date>\d{8}_\d{2}h).detided.symphonie_bobshelf.nc")
        super().__init__(loader)



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Interpolate SSH from MIT/GCM model
==================================
"""
import logging
import time

import dask.array as da
import numpy as np
import pyinterp
import pyinterp.geodetic
import xarray as xr

from .. import data_handler

LOGGER = logging.getLogger(__name__)


class MITGCM(data_handler.IrregularGridHandler):
    def __init__(self, grid_path: str, eta_path: str):
        loader = MITGCM.ZarrLoader(grid_path, eta_path)
        super().__init__(loader)

    class ZarrLoader(data_handler.DatasetLoader):
        def __init__(self, grid_path: str, eta_path: str):
            dataset = xr.merge(
                [xr.open_zarr(grid_path),
                 xr.open_zarr(eta_path)]).rename({
                     "XC": "lon",
                     "YC": "lat",
                     "Eta": "ssh"
                 })
            self.dataset = dataset[["ssh"]]
            self.dataset.dtime.load()
            self.time_delta = self._calculate_time_delta(self.dataset.dtime)

        def load_dataset(self, first_date: np.datetime64,
                         last_date: np.datetime64):
            first_date = self._shift_date(first_date.astype("datetime64[ns]"),
                                          -1, self.time_delta)
            last_date = self._shift_date(last_date.astype("datetime64[ns]"), 1,
                                         self.time_delta)

            if first_date < self.dataset.dtime[
                    0] or last_date > self.dataset.dtime[-1]:
                raise IndexError(
                    f"period [{first_date}, {last_date}] is out of range: "
                    f"[{self.dataset.dtime[0]}, {self.dataset.dtime[-1]}]")

            # Mask for selecting data covering the time period provided.
            mask = (self.dataset.dtime.data >=
                    first_date) & (self.dataset.dtime.data <= last_date)
            return self.dataset.isel(time=np.argwhere(mask).squeeze())

    @staticmethod
    def _spatial_interp(
        z_model: da.Array,
        x_model: da.Array,
        y_model: da.Array,
        x_sat: np.ndarray,
        y_sat: np.ndarray,
    ):
        mesh = pyinterp.RTree(dtype=np.dtype("float32"))
        x, y, z = (), (), ()

        start_time = time.time()

        for face in range(13):
            x_face = x_model[face, :].compute()
            y_face = y_model[face, :].compute()

            # We test if the face covers the satellite positions.
            ix0, ix1 = x_face.min(), x_face.max()
            iy0, iy1 = y_face.min(), y_face.max()

            box = pyinterp.geodetic.Box2D(pyinterp.geodetic.Point2D(ix0, iy0),
                                          pyinterp.geodetic.Point2D(ix1, iy1))
            mask = box.covered_by(x_sat, y_sat)
            if not np.any(mask == 1):
                continue
            del box, mask

            # The undefined values are filtered
            z_face = z_model[face, :].compute()
            defined = ~np.isnan(z_face)
            x += (x_face[defined].flatten(), )
            y += (y_face[defined].flatten(), )
            z += (z_face[defined].flatten(), )

        # The tree is built and the interpolation is calculated
        x = np.concatenate(x)
        y = np.concatenate(y)
        coordinates = np.vstack((x, y)).T
        del x, y

        z = np.concatenate(z)
        LOGGER.debug(
            "loaded %d MB in %.2fs",
            (coordinates.nbytes + z.nbytes) // 1024**2,
            time.time() - start_time,
        )
        start_time = time.time()
        mesh.packing(coordinates, z)
        LOGGER.debug("mesh build in %.2fs", time.time() - start_time)

        del coordinates, z

        start_time = time.time()
        z, _ = mesh.radial_basis_function(
            np.vstack((x_sat, y_sat)).T.astype("float32"),
            within=True,
            k=11,
            radius=55000,
            rbf="thin_plate",
            num_threads=1,
        )
        LOGGER.debug("interpolation done in %.2fs", time.time() - start_time)
        del mesh
        return z.astype("float32")



import os
import pickle
import numpy as np
import swot_simulator
import swot_simulator.error.altimeter
import swot_simulator.error.baseline_dilation
import swot_simulator.error.karin
import swot_simulator.error.roll_phase
import swot_simulator.error.timing
import swot_simulator.error.wet_troposphere
import swot_simulator.random_signal
import swot_simulator.settings

ROOT = os.path.dirname(os.path.abspath(__file__))

ERROR_SPECTRUM = swot_simulator.DATA.joinpath("error_spectrum.nc")


def load_parameters():
    ephemeris = swot_simulator.DATA.joinpath(
        "ephemeris_calval_june2015_ell.txt")
    karin_noise = swot_simulator.DATA.joinpath("karin_noise_v2.nc")
    return swot_simulator.settings.Parameters(
        dict(ephemeris=str(ephemeris),
             nadir=True,
             swath=True,
             error_spectrum=str(ERROR_SPECTRUM),
             karin_noise=str(karin_noise),
             cycle_duration=0.99349,
             height=857244))


def load_data(error=None):
    with open(os.path.join(ROOT, "data", "errors.bin"), "rb") as stream:
        (x_al, x_ac, curvilinear_distance, cycle_number,
         errors) = pickle.load(stream)
    return (x_al, x_ac, curvilinear_distance, cycle_number,
            errors[error] if error is not None else errors)


def test_error_karin():
    parameters = load_parameters()
    (x_al, x_ac, _curvilinear_distance, _cycle_number,
     _signal) = load_data('karin')
    error = swot_simulator.error.karin.Karin(parameters)
    swh = np.array([
        0,
    ] * len(x_ac))

    generated1 = error.generate(0, x_al, x_ac, swh)
    generated2 = error.generate(0, x_al, x_ac, swh)
    assert np.all(generated1['simulated_error_karin'] ==
                  generated2['simulated_error_karin'])
    generated2 = error.generate(1, x_al, x_ac, swh)
    assert np.all(generated1['simulated_error_karin'] !=
                  generated2['simulated_error_karin'])


def test_baseline_dilation():
    parameters = load_parameters()
    error_spectrum = swot_simulator.random_signal.read_file_instr(
        str(ERROR_SPECTRUM), 2.0)

    (x_al, x_ac, _, _, _) = load_data('baseline_dilation')
    error = swot_simulator.error.baseline_dilation.BaselineDilation(
        parameters, error_spectrum['dilationPSD'].data,
        error_spectrum['spatial_frequency'].data)
    generated = error.generate(x_al, x_ac)
    assert generated['simulated_error_baseline_dilation'].mean() < 1


def test_roll_phase():
    parameters = load_parameters()
    (x_al, x_ac, _, _, _) = load_data()
    error_spectrum = swot_simulator.random_signal.read_file_instr(
        str(ERROR_SPECTRUM), 2.0)

    error = swot_simulator.error.roll_phase.RollPhase(
        parameters, error_spectrum['rollPSD'].data,
        error_spectrum['gyroPSD'].data, error_spectrum['phasePSD'].data,
        error_spectrum['spatial_frequency'].data)
    generated = error.generate(x_al, x_ac)  # type: ignore
    assert generated['simulated_error_phase'].mean() < 1
    assert generated['simulated_error_roll'].mean() < 1


def test_timing():
    parameters = load_parameters()
    error_spectrum = swot_simulator.random_signal.read_file_instr(
        swot_simulator.DATA.joinpath("error_spectrum.nc"), 2.0)

    (x_al, x_ac, _, _, _) = load_data('timing')
    error = swot_simulator.error.timing.Timing(
        parameters, error_spectrum['timingPSD'].data,
        error_spectrum['spatial_frequency'].data)
    generated = error.generate(x_al, x_ac)
    assert generated['simulated_error_timing'].mean() < 1


def test_wet_troposphere():
    parameters = load_parameters()
    (x_al, x_ac, _, _, _) = load_data()

    parameters.nbeam = 1
    error = swot_simulator.error.wet_troposphere.WetTroposphere(parameters)
    generated = error.generate(x_al, x_ac)

    # assert abs(wt.data - expected['wt']['one']['wt']).mean() < 1e-12
    assert generated['simulated_error_troposphere'].mean() < 1

    # assert abs(wt_nadir.data -
    #            expected['wt_nadir']['one']['wt']).mean() < 1e-12
    assert generated['simulated_error_troposphere_nadir'].mean() < 1

    parameters.nbeam = 2
    error = swot_simulator.error.wet_troposphere.WetTroposphere(parameters)
    generated = error.generate(x_al, x_ac)

    # assert abs(wt.data - expected['wt']['two']['wt']).mean() < 1e-12
    assert generated['simulated_error_troposphere'].mean() < 1

    # assert abs(wt_nadir.data -
    #            expected['wt_nadir']['two']['wt']).mean() < 1e-12
    assert generated['simulated_error_troposphere_nadir'].mean() < 1


def test_altimeter():
    parameters = load_parameters()
    (x_al, _, _, _, _) = load_data('altimeter')

    error = swot_simulator.error.altimeter.Altimeter(parameters)
    generated = error.generate(x_al)
    assert generated['simulated_error_altimeter'].mean() < 1



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Plug-in handlers
----------------
"""
import numpy as np


class Interface:
    """Interface of a plugin"""
    @classmethod
    def interpolate(cls, lon: np.ndarray, lat: np.ndarray,
                    dates: np.ndarray) -> np.ndarray:
        """Interpolate the geophysical field for the given coordinates"""
        raise RuntimeError("You must register a plugin")


class Puppet(Interface):
    """Interpolation routine used for testing."""
    def interpolate(self, lon, _lat, _time):
        """Returns an array filled with zeros"""
        return np.full_like(lon, 0)


class Plugin:
    """Plug-in to interpolate a geophysical field"""
    def __init__(self):
        self.plugin = Interface()

    def interpolate(self, lon: np.ndarray, lat: np.ndarray,
                    time: np.ndarray) -> np.ndarray:
        """Interpolate the geophysical field for the given coordinates."""
        return self.plugin.interpolate(lon, lat, time)

    @classmethod
    def register(cls, plugin):
        """Register the user plugin"""
        if not isinstance(plugin, Interface):
            raise TypeError("plugin must be a sub-class of "
                            f"{Interface.__class__.__name__}")  # type: ignore
        result = cls()
        result.plugin = plugin
        return result



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Orbit Propagator
----------------
"""
from typing import Dict, Iterator, NamedTuple, Optional, TextIO, Tuple
import logging
import warnings
import numpy as np
from . import math
from . import settings
from . import VOLUMETRIC_MEAN_RADIUS

LOGGER = logging.getLogger(__name__)


def load_ephemeris(
    stream: TextIO,
    cols: Optional[Tuple[int, int,
                         int]] = None) -> Tuple[Dict[str, float], Tuple]:
    """Loads a tabular file describing a satellite orbit.

    Args:
        stream (typing.TextIO): Stream to the text file to be processed.
        cols (tuple): A tuple of 3 elements describing the columns
            representing the longitude, latitude and number of seconds elapsed
            since the beginning of the orbit.

    Returns:
        tuple: A tuple of 2 elements containing the properties of the orbit
        and the positions of the satellite.
    """
    if cols is None:
        cols = (1, 2, 0)
    LOGGER.info("loading ephemeris: %s", stream.name)

    comments = []
    data = []
    for item in stream:
        item = item.strip()
        if item.startswith("#"):
            comments.append(item[1:])
        else:
            data.append(list(float(value) for value in item.split()))
    data = np.asarray(data, dtype=np.float64)

    def to_dict(comments) -> Dict[str, float]:
        """Returns a dictionary describing the parameters of the orbit."""
        result = dict()
        for item in comments:
            key, value = item.split("=")
            result[key.strip()] = float(value)
        return result

    return to_dict(comments), tuple(data[:, item] for item in cols)


def interpolate(lon: np.ndarray, lat: np.ndarray, time: np.ndarray) -> Tuple:
    """Interpolate the given orbit at high resolution (0.5 seconds)

    Args:
        lon (numpy.ndarray): Longitudes (in degrees)
        lat (numpy.ndarray): Latitudes (in degrees)
        time (np.ndarray): Date of the positions (in seconds).

    Returns:
        tuple: lon, lat, and time interpolated
    """
    x, y, z = math.spher2cart(lon, lat)
    time_hr = np.arange(time[0], time[-1], 0.5, dtype=time.dtype)

    x = np.interp(time_hr, time, x)
    y = np.interp(time_hr, time, y)
    z = np.interp(time_hr, time, z)

    lon, lat = math.cart2spher(x, y, z)

    return lon, lat, time_hr


def rearrange_orbit(
        cycle_duration: float, lon: np.ndarray, lat: np.ndarray,
        time: np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Rearrange orbit starting from pass 1

    Detect the beginning of pass 1 in the ephemeris. By definition, it is
    the first passage at southernmost latitude.

    Args:
        cycle_duration (float): Cycle time in seconds.
        lon (numpy.ndarray): Longitudes (in degrees)
        lat (numpy.ndarray): Latitudes (in degrees)
        time (np.ndarray): Date of the positions (in seconds).

    Returns:
        tuple: lon, lat, and time rearranged.
    """
    dy = np.roll(lat, 1) - lat
    indexes = np.where((dy < 0) & (np.roll(dy, 1) >= 0))[0]

    # Shift coordinates, so that the first point of the orbit is the beginning
    # of pass 1
    shift = indexes[-1]

    lon = np.hstack([lon[shift:], lon[:shift]])
    lat = np.hstack([lat[shift:], lat[:shift]])
    time = np.hstack([time[shift:], time[:shift]])
    time = (time - time[0]) % cycle_duration
    if np.any(time < 0):
        warnings.warn('there are negative times in your orbit', RuntimeWarning)
    return lon, lat, time


def calculate_pass_time(lat: np.ndarray, time: np.ndarray) -> np.ndarray:
    """Compute the initial time of each pass

    Args:
        lat (numpy.ndarray): Latitudes (in degrees)
        time (np.ndarray): Date of the latitudes (in seconds).

    Returns:
        numpy.ndarray: Start date of half-orbits.
    """
    dy = np.roll(lat, 1) - lat
    indexes = np.where(((dy < 0) & (np.roll(dy, 1) >= 0))
                       | ((dy > 0)
                          & (np.roll(dy, 1) <= 0)))[0]
    # The duration of the first pass is zero.
    indexes[0] = 0
    return time[indexes]


def select_box(
        box: math.Box, lon: np.ndarray, lat: np.ndarray, time: np.ndarray
) -> Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
    """Selects the orbit in the defined box.

    Args:
        box (math.Box): Geographical selection
        lon (numpy.ndarray): Longitudes (in degrees)
        lat (numpy.ndarray): Latitudes (in degrees)
        time (np.ndarray): Date of the positions (in seconds).

    Returns:
        tuple: the selected positions and associated dates.
    """
    mask = box.within(lon, lat)
    x_al = math.curvilinear_distance(lon, lat, VOLUMETRIC_MEAN_RADIUS)
    return lon[mask], lat[mask], time[mask], x_al[mask]


class Orbit:
    """Properties of one orbit

    Args:
        height (float): Satellite height (in m)
        lat (numpy.ndarray): Latitudes (in degrees)
        lon (numpy.ndarray): Longitudes (in degrees)
        pass_time (np.ndarray): Start date of half-orbits.
        time (np.ndarray): Date of the positions (in seconds).
        x_al (np.ndarray): Along track distance
        curvilinear_distance (float): Distance covered by the satellite during
            a complete cycle.
        shift_time (float): Time shift to be applied.
    """
    def __init__(self, height: float, lat: np.ndarray, lon: np.ndarray,
                 pass_time: np.ndarray, time: np.ndarray, x_al: np.ndarray,
                 curvilinear_distance: float,
                 shift_time: Optional[np.timedelta64]):
        self.height = height
        self.lat = lat
        self.lon = lon
        self.pass_time = pass_time
        self.shift_time = shift_time
        self.time = time
        self.x_al = x_al
        self.curvilinear_distance = curvilinear_distance

    def cycle_duration(self) -> np.timedelta64:
        """Get the cycle duration"""
        return self.time[-1]

    def orbit_duration(self) -> np.timedelta64:
        """Get the orbit duration
        """
        duration = self.cycle_duration().astype(
            "timedelta64[us]") / np.timedelta64(
                int(self.passes_per_cycle() // 2), 'us')
        return np.timedelta64(int(duration), "us")

    def passes_per_cycle(self) -> int:
        """Get the number of passes per cycle"""
        return len(self.pass_time)

    def pass_shift(self, number: int) -> np.timedelta64:
        """Get the time offset between the first measurement and the last
        measurement of the track.

        Args:
            number (int): track number (must be in [1, passes_per_cycle()])

        Returns:
            numpy.datetime64: time offset
        """
        passes_per_cycle = self.passes_per_cycle()
        if number < 1 or number > passes_per_cycle:
            raise ValueError(f"number must be in [1, {passes_per_cycle}]")
        if number == passes_per_cycle:
            indexes = np.where(self.time >= self.pass_time[-1])[0]
        else:
            indexes = np.where((self.time >= self.pass_time[number - 1])
                               & (self.time < self.pass_time[number]))[0]
        return np.sum(np.diff(self.time[indexes]))

    def pass_duration(self, number: int) -> np.timedelta64:
        """Get the duration of a given pass.

        Args:
            number (int): track number (must be in [1, passes_per_cycle()])

        Returns:
            numpy.datetime64: track duration
        """
        passes_per_cycle = self.passes_per_cycle()
        if number < 1 or number > passes_per_cycle:
            raise ValueError(f"number must be in [1, {passes_per_cycle}]")
        if number == passes_per_cycle:
            return self.time[-1] - self.pass_time[-1] + self.time[
                1] - self.time[0]
        return self.pass_time[number] - self.pass_time[number - 1]

    def decode_absolute_pass_number(self, number: int) -> Tuple[int, int]:
        """Calculate the cycle and pass number from a given absolute pass
        number.

        Args:
            number (int): absolute pass number

        Returns:
            tuple: cycle and pass number
        """
        number -= 1
        return (int(number / self.passes_per_cycle()) + 1,
                (number % self.passes_per_cycle()) + 1)

    def encode_absolute_pass_number(self, cycle_number: int,
                                    pass_number: int) -> int:
        """Calculate the absolute pass number for a given half-orbit.

        Args:
            cycle_number (int): Cycle number
            pass_number (int): Pass number

        Returns:
            int: Absolute pass number
        """
        passes_per_cycle = self.passes_per_cycle()
        if not 1 <= pass_number <= passes_per_cycle:
            raise ValueError(f"pass_number must be in [1, {passes_per_cycle}")
        return (cycle_number - 1) * self.passes_per_cycle() + pass_number

    def delta_t(self) -> np.timedelta64:
        """Returns the average time difference between two measurements.

        Returns:
            int: average time difference
        """
        return np.diff(self.time).mean()

    def iterate(
        self,
        first_date: Optional[np.datetime64] = None,
        last_date: Optional[np.datetime64] = None,
        absolute_pass_number: int = 1
    ) -> Iterator[Tuple[int, int, np.datetime64]]:
        """Obtain all half-orbits within the defined time interval.

        Args:
            first_date (numpy.datetime64): First date of the period to be
                considered.
            last_date (numpy.datetime64): Last date of the period to be
                considered.
            absolute_pass_number (int, optional): Absolute number of the first
                pass to be returned.

        Returns:
            iterator: An iterator for all passes in the interval pointing to
            the cycle number, pass number and start date of the half-orbit.
        """
        date = first_date or np.datetime64("now")
        last_date = last_date or date + self.cycle_duration()
        while date <= last_date:
            cycle_number, pass_number = self.decode_absolute_pass_number(
                absolute_pass_number)

            yield cycle_number, pass_number, date

            # Shift the date of the duration of the generated pass
            date += self.pass_duration(pass_number)

            # Update of the number of the next pass to be generated
            absolute_pass_number += 1
        return StopIteration


class EquatorCoordinates(NamedTuple):
    """Coordinates of the satellite at the equator."""
    #: Longitude
    longitude: float
    #: Product dataset name
    time: np.datetime64


class Pass:
    """Handle one pass

    Args:
        lat_nadir (np.ndarray): Latitudes at nadir
        lat (np.ndarray): Latitudes of the swath
        lon_nadir (np.ndarray): Longitudes at nadir
        lon (np.ndarray): Longitudes of the swath
        time (np.ndarray): Time of measurements
        x_ac (np.ndarray): Across track distance.
        x_al (np.ndarray): Along track distance.
        equator_coordinate (EquatorCoordinates): Coordinates of the satellite at
            the equator.
        requirement_bounds: Limits of swath requirements. Measurements outside
            the span will be set to NaN.
    """
    def __init__(self,
                 lat_nadir: np.ndarray,
                 lat: np.ndarray,
                 lon_nadir: np.ndarray,
                 lon: np.ndarray,
                 time: np.ndarray,
                 x_ac: np.ndarray,
                 x_al: np.ndarray,
                 equator_coordinates: EquatorCoordinates,
                 requirement_bounds: Optional[Tuple[float, float]] = None):
        self.lat_nadir = lat_nadir
        self.lat = lat
        self.lon_nadir = lon_nadir
        self.lon = lon
        self.requirement_bounds = requirement_bounds
        self.timedelta = time
        self._date = None
        self.x_ac = x_ac
        self.x_al = x_al
        self._equator_coordinates = equator_coordinates

    @property
    def time(self):
        """Time of positions"""
        assert self._date is not None
        return self._date + (self.timedelta - self.timedelta[0])

    @property
    def time_eq(self):
        """Time at the equator"""
        assert self._date is not None
        return self._date + (self._equator_coordinates.time -
                             self.timedelta[0])

    @property
    def lon_eq(self):
        """Equator longitude"""
        return self._equator_coordinates.longitude

    def set_simulated_date(self, date: np.datetime64) -> None:
        """Set time of positions"""
        assert self._date is None
        self._date = date

    def mask(self) -> np.ndarray:
        """Obtain a mask to set NaN values outside the mission
        requirements."""
        if self.requirement_bounds is not None:
            valid = np.full_like(self.x_ac, np.nan)
            valid[(np.abs(self.x_ac) >= self.requirement_bounds[0])
                  & (np.abs(self.x_ac) <= self.requirement_bounds[1])] = 1
            along_track = np.full(self.lon_nadir.shape, 1, dtype=np.float64)
            return along_track[:, np.newaxis] * valid
        return np.full(self.lon.shape, 1, dtype=np.float64)


def calculate_orbit(parameters: settings.Parameters,
                    ephemeris: TextIO) -> Orbit:
    """Computes the orbit nadir on a subdomain.

    The path of the satellite is given by the orbit file and the subdomain
    corresponds to the one in the model. Note that a subdomain can be manually
    added in the parameters file.

    Args:
        parameters (settings.Parameters): Simulation parameters.
        ephemeris (typing.TextIO): Stream to the ephemeris CSV file to be
            processed.

    Returns:
        Orbit: The orbit's loaded into memory.
    """
    properties, (lon, lat,
                 time) = load_ephemeris(ephemeris,
                                        cols=parameters.ephemeris_cols)

    if np.any(lon > 360) or np.any(np.abs(lat) > 90):
        raise RuntimeError("The definition of ephemeris is incorrect. Is the "
                           "'ephemeris_cols' parameter well defined?")

    # Put longitude in [-180, 180[
    lon = math.normalize_longitude(lon)

    for key, value in properties.items():
        if key == 'cycle_duration':
            parameters.cycle_duration = value
        elif key == 'height':
            parameters.height = value
        else:
            raise RuntimeError(f"The {key!r} parameter defined in the "
                               "ephemeris is not handled.")

    if parameters.cycle_duration is None:
        raise RuntimeError("The cycle duration is not defined either in the "
                           "ephemeris file or in the simulator parameters.")

    if parameters.height is None:
        raise RuntimeError(
            "The satellite altitude is not defined either in "
            "the ephemeris file or in the simulator parameters.")

    # If orbit is at low resolution, interpolate the orbit provided
    if np.mean(np.diff(time)) > 0.5:
        lon, lat, time = interpolate(lon, lat, time)

    # Cut orbit if more than an orbit cycle is provided
    indexes = np.where(time < parameters.cycle_duration * 86400)[0]
    lon = lon[indexes]
    lat = lat[indexes]
    time = time[indexes]

    # Get cycle period.
    cycle_duration = time[-1] + time[1] - time[0]

    # shift time if the user needs to shift the time of the orbit
    if parameters.shift_time is not None:
        indexes = np.where(time >= parameters.shift_time)[0]
        lon = np.hstack([lon[indexes[0]:], lon[:indexes[0]]])
        lat = np.hstack([lat[indexes[0]:], lat[:indexes[0]]])
    del indexes

    if parameters.shift_lon is not None:
        lon = math.normalize_longitude(lon + parameters.shift_lon)

    # Rearrange orbit starting from pass 1
    lon, lat, time = rearrange_orbit(cycle_duration, lon, lat, time)

    # Calculates the along track distance
    distance = math.curvilinear_distance(lon, lat, VOLUMETRIC_MEAN_RADIUS)

    # Interpolate the final orbit according the given along track resolution
    x_al = np.arange(distance[0],
                     distance[-2],
                     parameters.delta_al,
                     dtype=distance.dtype)
    # In order to avoid any problem of calculation of dates on real numbers,
    # we manipulate dates encoded on integers from this point.
    time = (np.interp(x_al, distance[:-1], time[:-1]) *
            1e6).astype("timedelta64[us]")

    x, y, z = math.spher2cart(lon, lat)
    x = np.interp(x_al, distance[:-1], x[:-1])
    y = np.interp(x_al, distance[:-1], y[:-1])
    z = np.interp(x_al, distance[:-1], z[:-1])
    lon, lat = math.cart2spher(x, y, z)

    return Orbit(parameters.height, lat, lon,
                 np.sort(calculate_pass_time(lat, time)), time, x_al,
                 distance[-1], parameters.shift_time)


def equator_properties(lon_nadir: np.ndarray, lat_nadir: np.ndarray,
                       time: np.ndarray) -> EquatorCoordinates:
    """Calculate the position of the satellite at the equator"""
    # Search the nearest point to the equator
    i1 = (np.abs(lat_nadir)).argmin()
    i0 = i1 - 1
    if lat_nadir[i0] * lat_nadir[i1] > 0:
        i0, i1 = i1, i1 + 1
    lon1 = lon_nadir[i0:i1 + 1]
    lat1 = lat_nadir[i0:i1 + 1]
    time = time[i0:i1 + 1]

    # Calculate the intersection point
    p11, p12 = np.stack(math.spher2cart(lon1, lat1)).T
    p21, p22 = np.stack(
        math.spher2cart(lon1, np.zeros((2, ), dtype=np.float64))).T

    # Calculate normal of the normal vectors for arc 1 and 2
    l = np.cross(np.cross(p11, p12), np.cross(p21, p22))

    # Calculate one of the solution of the intersection point (the other is -i1)
    i1 = l / np.linalg.norm(l)
    lon_eq, _ = math.cart2spher(i1[:1], i1[1:2], i1[2:])
    if not lon1.min() <= lon_eq <= lon1.max():
        i1 = -i1
        lon_eq, _ = math.cart2spher(i1[:1], i1[1:2], i1[2:])

    # Calculate the distance along the track
    lon1 = np.insert(lon1, 1, lon_eq)
    lat1 = np.insert(lat1, 1, 0)
    x_al = math.curvilinear_distance(lon1, lat1, VOLUMETRIC_MEAN_RADIUS)

    # Pop the along track distance at the equator
    x_eq = x_al[1]
    x_al = np.delete(x_al, 1)

    # Finally interpolate the time ot the equator
    time_eq = np.interp(x_eq, x_al, time.astype("int64")).astype(time.dtype)
    return EquatorCoordinates(lon_eq[0], time_eq)


def calculate_pass(pass_number: int, orbit: Orbit,
                   parameters: settings.Parameters) -> Optional[Pass]:
    """Get the properties of an half-orbit

    Args:
        pass_number (int): Pass number
        orbit (Orbit): Orbit describing the pass to be calculated.
        parameters (settings.Parameters): Simulation parameters.

    Returns:
        Pass, optional: The properties of the pass or None, if the pas is
        missing in the selected area.
    """
    index = pass_number - 1
    # Selected indexes corresponding to the current pass
    if index == len(orbit.pass_time) - 1:
        indexes = np.where(orbit.time >= orbit.pass_time[-1])[0]
    else:
        indexes = np.where((orbit.time >= orbit.pass_time[index])
                           & (orbit.time < orbit.pass_time[index + 1]))[0]

    if len(indexes) < 5:
        return None

    lon_nadir = orbit.lon[indexes]
    lat_nadir = orbit.lat[indexes]
    time = orbit.time[indexes]
    x_al = orbit.x_al[indexes]

    equator_coordinates = equator_properties(lon_nadir, lat_nadir, time)

    # Selects the orbit in the defined box
    mask = parameters.box.within(lon_nadir, lat_nadir)
    if np.all(~mask):
        return None
    lon_nadir = lon_nadir[mask]
    lat_nadir = lat_nadir[mask]
    time = time[mask]
    x_al = x_al[mask]

    # Compute accross track distances from nadir
    # Number of points in half of the swath
    half_swath = int((parameters.half_swath - parameters.half_gap) /
                     parameters.delta_ac) + 1
    x_ac = np.arange(half_swath) * parameters.delta_al + parameters.half_gap
    x_ac = np.hstack((-np.flip(x_ac), x_ac))

    location = np.ascontiguousarray(
        np.vstack(math.spher2cart(lon_nadir, lat_nadir)).T)
    satellite_direction = math.satellite_direction(location)
    lon, lat = math.calculate_swath(parameters.delta_ac, parameters.half_gap,
                                    half_swath, VOLUMETRIC_MEAN_RADIUS,
                                    location, satellite_direction)

    return Pass(lat_nadir,
                lat,
                lon_nadir,
                lon,
                time,
                x_ac,
                x_al,
                equator_coordinates,
                requirement_bounds=parameters.requirement_bounds)



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Altimeter instrument error
--------------------------
"""
from typing import Dict
import numpy as np

from .. import random_signal
from .. import settings


class Altimeter:
    """Altimeter instrument error.

    Args:
        parameters (settings.Parameters): Simulation settings
    """
    def __init__(self, parameters: settings.Parameters) -> None:
        # Store the generation parameters of the random signal.
        self.delta_al = 2 * parameters.delta_al
        self.rng = parameters.rng()
        self.len_repeat = parameters.len_repeat
        # Define the sepctrum of the nadir instrument error
        freq = np.arange(1 / 3000, 1 / self.delta_al, 1 / 3000)
        psd = 8 + 1.05 * 1e-4 * freq**(-2.2)
        psd[freq < 0.00023627939582672978] = 1e4

        # Convert spectrum in m2/cy
        self.psd = psd * 1e-4
        self.freq = freq

    def generate(self, x_al: np.array) -> Dict[str, np.ndarray]:
        """Generate altimeter instrument error.

        Args:
            x_al (numpy.ndarray): Along track distance

        Returns:
            dict: variable name and errors simulated.
        """
        # Compute random noise of 10**2 cm**2/(km/cycle)
        # Compute the correspond error on the nadir in m
        error = random_signal.gen_signal_1d(self.freq,
                                            self.psd,
                                            x_al,
                                            rng=self.rng,
                                            fmin=1 / self.len_repeat,
                                            fmax=1 / self.delta_al,
                                            alpha=10)
        return {"simulated_error_altimeter": error}



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Karin noise
-----------
"""
from typing import Dict
import numpy as np

from .. import random_signal
from .. import settings


class Karin:
    """Karin instrumental error computed from random realization

    Args:
        parameters (settings.Parameters): Simulation settings
    """
    def __init__(self, parameters: settings.Parameters) -> None:
        assert parameters.karin_noise is not None

        # Store the generation parameters of the random signal.
        self.hsdt, self.x_ac, self.swh = random_signal.read_file_karin(
            parameters.karin_noise)

        self.size_grid = (parameters.delta_ac * parameters.delta_al)**0.5

        # Hack for unsmoothed products at high resolution
        if self.size_grid < 1:
            self.size_grid *= 8 / (40**.5)

    def generate(self, seed: int, x_al: np.ndarray, x_ac: np.ndarray,
                 swh: np.ndarray) -> Dict[str, np.ndarray]:
        """Generate the karin noise

        Args:
            seed (int): Random seed used to initialize the pseudo-random
                number generator.
            x_al (numpy.ndarray): Along track distance
            x_ac (numpy.ndarray): Across track distance
            swh (numpy.ndarray): Significant wave height. Used to modulate
                instrumental noise as a function of sea state.

        Returns:
            dict: variable name and errors simulated.
        """
        num_pixels = x_ac.shape[0]
        num_lines = x_al.shape[0]

        # Generate random noise for left and right part of the mast
        rng = np.random.default_rng(seed=seed)
        a_karin = rng.normal(0, 1, (num_lines, num_pixels))

        # Formula of karin noise as a function of x_ac (smile shape)
        sigma_karin = random_signal.interpolate_file_karin(
            swh, x_ac, self.hsdt, self.x_ac, self.swh) / self.size_grid

        # Compute random karin error
        return {"simulated_error_karin": sigma_karin * a_karin}



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Plug-in to interpolate SWH
--------------------------
"""
from .ww3 import WW3



# Copyright (c) 2021 CNES/JPL
#
# All rights reserved. Use of this source code is governed by a
# BSD-style license that can be found in the LICENSE file.
"""
Plug-in to interpolate SSH
--------------------------
"""
from .aviso import AVISO
from .hycom import HYCOM
from .mitgcm import MITGCM
from .mitgcm_ww3 import MITGCM_WW3
from .nemo import NEMO
from .symphonie import SYMPHONIE
